import re
from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import TYPE_CHECKING, Iterable, List, Tuple

from ldb.cli_utils import add_data_obj_params
from ldb.core import get_ldb_instance
from ldb.pull import pull

if TYPE_CHECKING:
    from argparse import _SubParsersAction

VERSION_RE = r"^v(\d+)$"


def pull_command(options: Namespace) -> None:
    ldb_dir = get_ldb_instance()
    paths, version = get_version_and_paths(options.paths)
    pull(
        ldb_dir,
        Path("."),
        paths,
        options.query_args,
        version,
        warn=False,
    )


def get_version_and_paths(paths: List[str]) -> Tuple[List[str], int]:
    if paths:
        version = get_version(paths[-1])
        if version != -1:
            paths = paths[:-1]
    else:
        version = -1
    return paths, version


def get_version(s: str) -> int:
    match = re.search(VERSION_RE, s)
    if match is not None:
        return int(match.group(1))
    return -1


def add_parser(
    subparsers: "_SubParsersAction[ArgumentParser]",
    parents: Iterable[ArgumentParser],
) -> None:
    help_msg = "Update annotation versions."
    desc = (
        "By default each annotation is updated to the newest version. If the "
        "final positional argument is a version identifier (e.g. v2), it will "
        "be used instead."
    )
    parser = subparsers.add_parser(
        "pull",
        parents=parents,
        help=help_msg,
        description=f"{help_msg} {desc}",
    )
    add_data_obj_params(parser, dest="query_args")
    parser.set_defaults(func=pull_command)



import getpass
import json
import os
import re
from dataclasses import dataclass
from datetime import date, datetime
from pathlib import Path
from typing import (
    Any,
    Collection,
    Dict,
    Iterable,
    List,
    Mapping,
    NamedTuple,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
)

import fsspec
from fsspec.spec import AbstractFileSystem
from fsspec.utils import get_protocol

from ldb.data_formats import Format
from ldb.exceptions import IndexingException, NotAStorageLocationError
from ldb.fs import posix_path as fsp
from ldb.fs.utils import cp_file_any_fs, has_protocol
from ldb.func_utils import apply_optional
from ldb.storage import StorageLocation, get_filesystem
from ldb.typing import JSONDecoded
from ldb.utils import (
    format_datetime,
    get_file_hash,
    get_filetype,
    get_first,
    get_fsspec_path_suffix,
    parse_datetime,
    timestamp_to_datetime,
    unique_id,
)

ENDING_DOUBLE_STAR_RE = r"(?:/+\*\*)+/*$"

AnnotationMeta = Dict[str, Union[str, int, None]]
DataObjectMeta = Dict[
    str,
    Union[str, None, Dict[str, Union[str, List[str], int, None]]],
]
DataToWrite = Tuple[Path, bytes, bool]
IndexingJob = Tuple["IndexingConfig", List[str]]
IndexingJobMapping = Dict[AbstractFileSystem, List[IndexingJob]]
FSPathsMapping = Dict[AbstractFileSystem, List[str]]


class FileSystemPath(NamedTuple):
    fs: AbstractFileSystem
    path: str


@dataclass
class IndexingConfig:
    save_data_object_path_info: bool


DEFAULT_CONFIG = IndexingConfig(save_data_object_path_info=True)
INDEXED_EPHEMERAL_CONFIG = IndexingConfig(save_data_object_path_info=False)


def expand_indexing_paths(
    paths: Iterable[str],
    storage_locations: Iterable[StorageLocation],
    default_format: bool = False,
) -> Dict[AbstractFileSystem, List[str]]:
    storage_locations = list(storage_locations)
    path_collections: Dict[AbstractFileSystem, Tuple[List[str], Set[str]]] = {}
    for indexing_path in paths:
        fs, paths_found = expand_single_indexing_path(
            indexing_path,
            storage_locations,
            default_format=default_format,
        )
        try:
            fs_paths, seen = path_collections[fs]
        except KeyError:
            fs_paths, seen = [], set()
            path_collections[fs] = fs_paths, seen
        for path in paths_found:
            if path not in seen:
                fs_paths.append(path)
                seen.add(path)
    return {fs: fs_paths for fs, (fs_paths, _) in path_collections.items()}


def expand_single_indexing_path(
    path: str,
    storage_locations: Collection[StorageLocation],
    default_format: bool = False,
) -> Tuple[AbstractFileSystem, List[str]]:
    """
    Get storage paths for indexing that match the glob, `path`.

    Because every file path under `path` is returned, any final "/**" does not
    change the result. First, `path` is expanded with any final "/**" removed
    and any matching files are included in the result. Corresponding data
    object file paths are added for matched annotation file paths and vice
    versa. Then everything under matched directories will be added to the
    result.

    The current implementation may result in some directory paths and some
    duplicate paths being included.
    """
    # TODO: make sure path is in a storage location if necessary
    protocol = get_protocol(path)
    fs_cls = fsspec.get_filesystem_class(protocol)
    path = fs_cls._strip_protocol(path)  # pylint: disable=protected-access
    fs = get_filesystem(path, protocol, storage_locations)
    if protocol == "file":
        path = os.path.abspath(path)
    if is_hidden_fsspec_path(path):
        return fs, []
    # "path/**", "path/**/**" or "path/**/" are treated the same as just "path"
    # so we strip off any "/**" or "/**/" at the end
    path = re.sub(ENDING_DOUBLE_STAR_RE, "", path)
    # find corresponding data object for annotation match and vice versa
    # for any files the expanded `path` glob matches
    path_match_globs = []
    for epath in fs.expand_path(path):
        if not is_hidden_fsspec_path(epath) and fs.isfile(epath):
            path_match_globs.append(epath)
            if default_format:
                # TODO: Check all extension levels (i.e. for abc.tar.gz use
                # .tar.gz and .gz)
                p_without_ext, ext = os.path.splitext(epath)
                if ext == ".json":
                    path_match_globs.append(p_without_ext)
                    path_match_globs.append(p_without_ext + ".*")
                else:
                    path_match_globs.append(p_without_ext + ".json")
    paths = (
        [i for p in path_match_globs for i in fs.glob(p)]
        if path_match_globs
        else []
    )
    if protocol not in ("http", "https"):
        # capture everything under any directories the `path` glob matches
        for epath in fs.expand_path(path, recursive=True):
            if not is_hidden_fsspec_path(epath) and fs.isfile(epath):
                paths.append(epath)
    return fs, paths


def is_hidden_fsspec_path(path: str) -> bool:
    return re.search(r"(?:/|^)\.(?!/|$)", path) is not None


def group_indexing_paths_by_type(
    fs: AbstractFileSystem,
    storage_paths: Iterable[str],
) -> Tuple[List[str], List[str]]:
    data_object_paths = []
    annotation_paths = []
    seen = set()
    for storage_path in storage_paths:
        if storage_path not in seen:
            seen.add(storage_path)
            if fs.isfile(storage_path):
                if storage_path.endswith(".json"):
                    annotation_paths.append(storage_path)
                else:
                    data_object_paths.append(storage_path)
    return data_object_paths, annotation_paths


def expand_dir_paths(
    paths: Iterable[str],
) -> Dict[AbstractFileSystem, List[str]]:
    for path in paths:
        path, num = re.subn(ENDING_DOUBLE_STAR_RE, "", path)
        if num:
            raise IndexingException(
                f"Paths passed with the {Format.INFER} format should only "
                "match directories, so globs with a final /** should not be "
                "used",
            )

    path_sets: Dict[AbstractFileSystem, Set[str]] = {}
    for path in paths:
        if not is_hidden_fsspec_path(path):
            fs = fsspec.filesystem(get_protocol(path))
            fs_paths = path_sets.setdefault(fs, set())
            for p in fs.expand_path(paths):
                if not is_hidden_fsspec_path(p) and fs.isdir(p):
                    fs_paths.add(p)
    path_lists: Dict[AbstractFileSystem, List[str]] = {}
    for fs, fs_path_set in path_sets.items():
        paths = sorted(fs_path_set)
        for i in range(len(paths) - 1):
            if paths[i + 1].startswith(paths[i]):
                raise IndexingException(
                    f"Paths passed with the {Format.INFER} format should "
                    "match non-overlapping directories. Found overlapping "
                    "directories:\n"
                    f"{paths[i]!r}\n"
                    f"{paths[i + 1]!r}",
                )
        path_lists[fs] = paths
    return path_lists


def create_storage_path(
    dest_fs: AbstractFileSystem,
    base_dir: str,
    source_fs: AbstractFileSystem,
    path: str,
) -> str:
    if source_fs.protocol == "file":
        # handle windows drive
        path = re.sub("^([A-Za-z]):", r"\1", path)
    path = dest_fs.sep.join(
        [base_dir] + path.lstrip(source_fs.sep).split(source_fs.sep),
    )
    return path


def copy_to_read_add_storage(
    data_object_paths: FSPathsMapping,
    annotation_paths: FSPathsMapping,
    read_add_location: StorageLocation,
    hashes: Mapping[AbstractFileSystem, Mapping[str, str]],
    strict_format: bool,
) -> Tuple[
    Dict[AbstractFileSystem, Dict[str, FileSystemPath]],
    Dict[AbstractFileSystem, Dict[str, FileSystemPath]],
]:
    dest_fs = fsspec.filesystem(
        read_add_location.protocol,
        **read_add_location.options,
    )
    read_add_path = (
        dest_fs._strip_protocol(  # pylint: disable=protected-access
            read_add_location.path,
        )
    )
    base_dir = dest_fs.sep.join(
        [
            *read_add_path.split("/"),
            "ldb-autoimport",
            date.today().isoformat(),
            unique_id(),
        ],
    )
    dest_fs.makedirs(base_dir, exist_ok=True)
    old_to_new_paths = {}
    old_to_new_annot_paths = {}
    for source_fs, paths in data_object_paths.items():
        fs_id = source_fs.protocol  # TODO use actual fs_id
        fs_base_dir = dest_fs.sep.join([base_dir, fs_id])

        fs_annotation_paths = set(annotation_paths.get(source_fs, []))

        fs_old_to_new_paths = {}
        fs_old_to_new_annot_paths = {}
        for path in paths:
            dest = create_storage_path(dest_fs, fs_base_dir, source_fs, path)
            annotation_dest = None
            annotation_path = data_object_path_to_annotation_path(path)
            if annotation_path in fs_annotation_paths:
                annotation_dest = create_storage_path(
                    dest_fs,
                    fs_base_dir,
                    source_fs,
                    annotation_path,
                )
            elif strict_format:
                continue
            try:
                dest_fs.makedirs(
                    dest_fs._parent(dest),  # pylint: disable=protected-access)
                    exist_ok=True,
                )
                if annotation_dest is not None:
                    cp_file_any_fs(
                        source_fs,
                        annotation_path,
                        dest_fs,
                        annotation_dest,
                    )
                cp_file_any_fs(source_fs, path, dest_fs, dest)
            except FileNotFoundError:
                # Use hash instead of preserving path if path is too long
                try:
                    data_object_hash = hashes[source_fs][path]
                except KeyError:
                    data_object_hash = get_file_hash(
                        source_fs,
                        path,
                    )
                dest = dest_fs.sep.join(
                    [
                        base_dir,
                        data_object_hash + get_fsspec_path_suffix(path),
                    ],
                )
                if annotation_dest is not None:
                    annotation_dest = data_object_path_to_annotation_path(dest)
                    cp_file_any_fs(
                        source_fs,
                        annotation_path,
                        dest_fs,
                        annotation_dest,
                    )
                cp_file_any_fs(source_fs, path, dest_fs, dest)
            fs_old_to_new_paths[path] = FileSystemPath(dest_fs, dest)
            if annotation_dest is not None:
                fs_old_to_new_annot_paths[annotation_path] = FileSystemPath(
                    dest_fs,
                    annotation_dest,
                )
        if fs_old_to_new_paths:
            old_to_new_paths[dest_fs] = fs_old_to_new_paths
        if fs_old_to_new_annot_paths:
            old_to_new_annot_paths[dest_fs] = fs_old_to_new_annot_paths
    return old_to_new_paths, old_to_new_annot_paths


def data_object_path_to_annotation_path(path: str) -> str:
    return os.path.splitext(path)[0] + ".json"


def separate_local_and_cloud_files(
    paths: FSPathsMapping,
) -> Tuple[FSPathsMapping, FSPathsMapping]:
    local = {}
    cloud = {}
    for fs, fs_paths in paths.items():
        if fs.protocol == "file":
            local[fs] = fs_paths
        else:
            cloud[fs] = fs_paths
    return local, cloud


def separate_storage_and_non_storage_files(
    fs_paths: FSPathsMapping,
    storage_locations: Sequence[StorageLocation],
) -> Tuple[FSPathsMapping, FSPathsMapping]:
    storage = {}
    non_storage = {}
    for fs, paths in fs_paths.items():
        matching_locs = filter_storage_locations(fs, storage_locations)
        if not matching_locs:
            non_storage[fs] = paths.copy()
        else:
            fs_storage = []
            fs_non_storage = []
            for path in paths:
                if in_storage_locations(path, matching_locs):
                    fs_storage.append(path)
                else:
                    fs_non_storage.append(path)
            if fs_storage:
                storage[fs] = fs_storage
            if fs_non_storage:
                non_storage[fs] = fs_non_storage
    return storage, non_storage


def separate_indexed_files(
    existing_hashes: Set[str],
    hashes: Mapping[AbstractFileSystem, Mapping[str, str]],
    fs_paths: FSPathsMapping,
) -> Tuple[FSPathsMapping, FSPathsMapping]:
    indexed = {}
    not_indexed = {}
    for fs, paths in fs_paths.items():
        fs_indexed = []
        fs_not_indexed = []
        for path in paths:
            try:
                hash_str: Optional[str] = hashes[fs][path]
            except KeyError:
                hash_str = None
            if hash_str is not None and hash_str in existing_hashes:
                fs_indexed.append(path)
            else:
                fs_not_indexed.append(path)
        if fs_indexed:
            indexed[fs] = fs_indexed
        if fs_not_indexed:
            not_indexed[fs] = fs_not_indexed
    return indexed, not_indexed


def validate_locations_in_storage(
    fs_paths: FSPathsMapping,
    storage_locations: Sequence[StorageLocation],
) -> None:
    for fs, paths in fs_paths.items():
        matching_locs = filter_storage_locations(fs, storage_locations)
        for path in paths:
            if not in_storage_locations(path, matching_locs):
                raise NotAStorageLocationError(
                    "Found file outside of configured storage locations: "
                    f"{path}",
                )


def filter_storage_locations(
    fs: AbstractFileSystem,
    storage_locations: Sequence[StorageLocation],
) -> List[StorageLocation]:
    # TODO use fs_id
    return [
        s for s in storage_locations if has_protocol(fs.protocol, s.protocol)
    ]


def in_storage_locations(
    path: str,
    storage_locations: Sequence[StorageLocation],
) -> bool:
    for loc in storage_locations:
        if fsp.isin(path, loc.path):
            return True
    return False


def datetime_fs_info(
    fs_info: Dict[str, Any],
    *keys: str,
) -> Optional[datetime]:
    return apply_optional(timestamp_to_datetime, get_first(fs_info, *keys))


def max_datetime_info(
    *datetimes: Union[datetime, str, None],
) -> Optional[datetime]:
    return max(
        (
            parse_datetime(d) if isinstance(d, str) else d
            for d in datetimes
            if d is not None
        ),
        default=None,
    )


def construct_data_object_meta(
    fs: AbstractFileSystem,
    path: str,
    prev_meta: Dict[str, Any],
    current_timestamp: str,
    tags: Collection[str] = (),
) -> DataObjectMeta:
    # TODO: create dataclass for data object meta
    fs_info = fs.info(path)

    last_indexed_by: str = getpass.getuser()
    atime = datetime_fs_info(fs_info, "atime", "accessed", "time")
    mtime = datetime_fs_info(fs_info, "mtime", "modified")
    ctime = datetime_fs_info(fs_info, "ctime", "created")
    size: int = fs_info.get("size", 0)
    mode: int = fs_info.get("mode", 0)
    uid: Optional[int] = fs_info.get("uid")
    gid: Optional[int] = fs_info.get("gid")
    filetype: str = get_filetype(path)

    if prev_meta:
        first_indexed = prev_meta["first_indexed"]
        new_tags = sorted(set(prev_meta["tags"]) | set(tags))
        alternate_paths = prev_meta["alternate_paths"]
        atime = max_datetime_info(atime, prev_meta["fs"]["atime"])
        mtime = max_datetime_info(mtime, prev_meta["fs"]["mtime"])
        ctime = max_datetime_info(ctime, prev_meta["fs"]["ctime"])
        filetype = filetype or prev_meta["type"]
    else:
        first_indexed = current_timestamp
        new_tags = sorted(set(tags))
        alternate_paths = []

    protocol: Union[str, List[str]] = fs.protocol
    path_info = {
        "fs_id": "",
        "protocol": protocol,
        "path": path,
    }
    if path_info not in alternate_paths:
        alternate_paths.append(path_info)
    return {
        "type": filetype,
        "first_indexed": first_indexed,
        "last_indexed": current_timestamp,
        "last_indexed_by": last_indexed_by,
        "tags": new_tags,  # type: ignore[dict-item]
        "alternate_paths": alternate_paths,
        "fs": {
            **path_info,
            "size": size,
            "mode": mode,
            "uid": uid,
            "gid": gid,
            "atime": apply_optional(format_datetime, atime),
            "mtime": apply_optional(format_datetime, mtime),
            "ctime": apply_optional(format_datetime, ctime),
        },
    }


def parse_annotation(raw_annotation: str) -> JSONDecoded:
    try:
        parsed_annotation: JSONDecoded = json.loads(raw_annotation)
    except Exception as exc:
        raise IndexingException(  # TODO: use more appropriate exception type
            "Unable to parse JSON annotation\n" f"{type(exc).__name__}: {exc}",
        ) from exc
    return parsed_annotation


def get_annotation_content(
    fs: AbstractFileSystem,
    path: str,
) -> JSONDecoded:
    try:
        with fs.open(path, "r") as file:
            raw_annotation = file.read()
        return parse_annotation(raw_annotation)
    except Exception as exc:
        raise IndexingException(
            f"Unable to read annotation file: {path}",
        ) from exc


def construct_annotation_meta(
    prev_annotation_meta: AnnotationMeta,
    current_timestamp: str,
    version: int,
    curr_mtime: Optional[datetime],
) -> AnnotationMeta:
    mtimes = []
    if prev_annotation_meta:
        prev_mtime: Optional[str] = prev_annotation_meta.get(  # type: ignore[assignment] # noqa: E501
            "mtime",
        )
        if prev_mtime is not None:
            mtimes.append(parse_datetime(prev_mtime))
        version = prev_annotation_meta["version"]  # type: ignore[assignment]
        first_indexed_time = prev_annotation_meta["first_indexed_time"]
    else:
        first_indexed_time = current_timestamp

    if curr_mtime is not None:
        mtimes.append(curr_mtime)

    mtime = format_datetime(max(mtimes)) if mtimes else None
    return {
        "version": version,
        "mtime": mtime,
        "first_indexed_time": first_indexed_time,
        "last_indexed_time": current_timestamp,
    }



import os
from contextlib import contextmanager
from copy import copy
from dataclasses import asdict, dataclass, field
from json import dump, load
from pathlib import Path
from typing import Collection, Dict, Generator, Iterable, List, Optional, Union

import fsspec
from fsspec.spec import AbstractFileSystem
from fsspec.utils import get_protocol

from ldb.exceptions import LDBException, StorageConfigurationError
from ldb.fs import posix_path as fsp
from ldb.path import Filename

FSOptions = Dict[str, Union[str, int, bool]]


@dataclass
class StorageLocation:
    path: str = ""
    protocol: str = "file"
    fs_id: str = ""
    read_access_verified: bool = False
    write_access_verified: bool = False
    read_and_add: bool = False
    options: FSOptions = field(default_factory=dict)


@dataclass
class StorageConfig:
    locations: List[StorageLocation] = field(default_factory=list)


def get_storage_locations(ldb_dir: Path) -> List[StorageLocation]:
    storage_path = ldb_dir / Filename.STORAGE
    if storage_path.is_file():
        return load_from_path(storage_path).locations
    return []


def get_containing_storage_location(
    path: str,
    protocol: str,
    storage_locations: Collection[StorageLocation],
) -> Optional[StorageLocation]:
    for loc in storage_locations:
        if loc.protocol == protocol and fsp.isin(path, loc.path):
            return loc
    return None


def get_filesystem(
    path: str,
    protocol: str,
    storage_locations: Collection[StorageLocation],
) -> AbstractFileSystem:
    storage_location = get_containing_storage_location(
        path,
        protocol,
        storage_locations,
    )
    fs_options = {} if storage_location is None else storage_location.options
    return fsspec.filesystem(protocol, **fs_options)


def load_from_path(path: Path) -> StorageConfig:
    with path.open() as file:
        config_dict = load(file)
    return StorageConfig(
        locations=[StorageLocation(**loc) for loc in config_dict["locations"]],
    )


def save_to_path(storage_config: StorageConfig, path: Path) -> None:
    config_dict = asdict(storage_config)
    with path.open("w") as file:
        dump(config_dict, file)


def create_storage_location(
    path: str = "",
    protocol: str = "",
    read_and_add: bool = False,
    options: Optional[FSOptions] = None,
) -> StorageLocation:
    if not protocol:
        protocol = get_protocol(path)
    fs = fsspec.filesystem(protocol)
    if protocol == "file":
        path = os.path.abspath(path)
    path = fs._strip_protocol(path)  # pylint: disable=protected-access
    return StorageLocation(
        path=path,
        protocol=protocol,
        read_access_verified=os.access(path, os.R_OK),
        write_access_verified=os.access(path, os.W_OK),
        read_and_add=read_and_add,
        options=options if options is not None else {},
    )


@contextmanager
def edit(path: Path) -> Generator[StorageConfig, None, None]:
    try:
        storage_config: StorageConfig = load_from_path(path)
    except FileNotFoundError:
        storage_config = StorageConfig()
    yield storage_config
    save_to_path(storage_config, path)


def add_storage(
    storage_config_filepath: Path,
    storage_location: StorageLocation,
    force: bool = False,
) -> None:
    storage_location = copy(storage_location)
    with edit(storage_config_filepath) as storage_config:
        new_locations = []
        children = []
        to_replace = None
        for loc in storage_config.locations:
            keep = True
            if loc.protocol == storage_location.protocol:
                old_path = loc.path
                new_path = storage_location.path
                if old_path == new_path:
                    if (
                        storage_location.read_and_add
                        and not loc.read_and_add
                        or storage_location.options != loc.options
                    ):
                        to_replace = loc
                        keep = False
                    else:
                        raise LDBException(
                            "The storage location "
                            f"{repr(storage_location.path)} already exists",
                        )
                elif fsp.isin(new_path, old_path):
                    raise LDBException(
                        f"{repr(storage_location.path)} is inside existing "
                        f"storage location {repr(loc.path)}",
                    )
                elif fsp.isin(old_path, new_path):
                    children.append(loc)
                    keep = False
            if keep:
                new_locations.append(loc)

        # When replacing any existing location, inherit any true
        # read_and_add setting
        to_replace_list = [to_replace] if to_replace is not None else children
        for loc in to_replace_list:
            if loc.read_and_add:
                storage_location.read_and_add = True

        if to_replace is not None:
            output = get_update_output(to_replace, storage_location)
        elif children:
            children_str = "\n".join(
                [f"  {repr(loc.path)}" for loc in children],
            )
            if not force:
                raise LDBException(
                    f"{repr(storage_location.path)} is a parent of "
                    f"existing storage locations:\n"
                    f"{children_str}\n"
                    "Use the --force option to replace them",
                )
            output = (
                "Added storage location {repr(storage_location.path)}\n"
                "Removed its children:\n"
                f"{children_str}"
            )
        else:
            output = f"Added storage location {repr(storage_location.path)}"

        new_locations.append(storage_location)
        validate_storage_locations(new_locations)
        storage_config.locations = new_locations
    print(output)


def get_update_output(old: StorageLocation, new: StorageLocation) -> str:
    old_dict = asdict(old)
    new_dict = asdict(new)
    updates = [(k, old_dict[k], new_dict[k]) for k in new_dict]
    update_str = "\n".join(
        [f"  {k}: {repr(o)} -> {repr(n)}" for k, o, n in updates if o != n],
    )
    return "Updated storage location " f"{repr(new.path)}:\n" f"{update_str}\n"


def validate_storage_locations(locations: Iterable[StorageLocation]) -> None:
    if sum(loc.read_and_add for loc in locations) > 1:
        raise StorageConfigurationError(
            "Only one storage location may be set as read-add",
        )



from typing import Any, Callable, overload


@overload
def apply_optional(func: Callable[[Any], Any], arg: None) -> None:
    ...


@overload
def apply_optional(func: Callable[[Any], Any], arg: Any) -> Any:
    ...


def apply_optional(func: Callable[[Any], Any], arg: Any) -> Any:
    if arg is None:
        return None
    return func(arg)



import os
import warnings
from pathlib import Path
from types import ModuleType
from typing import Iterable, Iterator, Set, Tuple

from ldb.exceptions import LDBException
from ldb.import_utils import import_path
from ldb.path import InstanceDir
from ldb.typing import JSONFuncMutableMapping

CUSTOM_FUNCTIONS_VAR = "CUSTOM_FUNCTIONS"


def load_user_functions(
    ldb_dir: Path,
    invalid_names: Iterable[str] = (),
) -> JSONFuncMutableMapping:
    invalid_names = set(invalid_names)
    user_functions: JSONFuncMutableMapping = {}
    try:
        for module, custom_functions in list(
            iter_user_function_modules(ldb_dir),
        ):
            validate_func_names(module, custom_functions, invalid_names)
            user_functions.update(custom_functions)
    except Exception as exc:
        raise LDBException(
            "Unable to load custom user functions: "
            f"{type(exc).__name__}:\n{exc}",
        ) from exc
    return user_functions


def iter_user_function_modules(
    ldb_dir: Path,
) -> Iterator[Tuple[ModuleType, JSONFuncMutableMapping]]:
    for path in sorted((ldb_dir / InstanceDir.USER_FUNCTIONS).glob("*.py")):
        module = import_path(os.fspath(path))
        try:
            custom_functions = getattr(module, CUSTOM_FUNCTIONS_VAR)
        except AttributeError:
            warnings.warn(
                f"Found user function module {module.__name__} "
                f"but no {CUSTOM_FUNCTIONS_VAR} variable: ",
            )
        else:
            yield module, custom_functions


def validate_func_names(
    module: ModuleType,
    func_names: Iterable[str],
    invalid_names: Set[str],
) -> None:
    for func_name in func_names:
        if func_name in invalid_names:
            raise ValueError(
                f"Invalid user function name: {func_name!r}\n"
                "A function with this name already exists.\n"
                f"(module_name={module.__name__!r}, "
                f"file_path={module.__file__!r})",
            )



import json
import os
import shutil
import subprocess
import tempfile
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from pathlib import Path
from typing import (
    Collection,
    List,
    Mapping,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
    Union,
    cast,
)

from funcy.objects import cached_property

from ldb.data_formats import INSTANTIATE_FORMATS, Format
from ldb.dataset import get_annotation
from ldb.exceptions import LDBException
from ldb.fs.utils import FSProtocol, first_protocol
from ldb.path import InstanceDir, WorkspacePath
from ldb.progress import get_progressbar
from ldb.storage import StorageLocation, get_filesystem, get_storage_locations
from ldb.typing import JSONDecoded, JSONObject
from ldb.utils import (
    DATA_OBJ_ID_PREFIX,
    get_hash_path,
    json_dumps,
    load_data_file,
    write_data_file,
)
from ldb.workspace import (
    collection_dir_to_object,
    ensure_path_is_empty_workspace,
)


class InstantiateResult(NamedTuple):
    data_object_paths: List[str]
    annotation_paths: List[str]
    num_data_objects: int
    num_annotations: int


def instantiate(
    ldb_dir: Path,
    workspace_path: Path,
    dest: Path,
    fmt: str = Format.BARE,
    force: bool = False,
    apply: Sequence[str] = (),
) -> InstantiateResult:
    fmt = INSTANTIATE_FORMATS[fmt]
    collection = collection_dir_to_object(
        workspace_path / WorkspacePath.COLLECTION,
    )

    # fail fast if workspace is not empty
    if dest.exists():
        ensure_path_is_empty_workspace(dest, force)
    dest.mkdir(exist_ok=True)

    tmp_dir_base = workspace_path / WorkspacePath.TMP
    tmp_dir_base.mkdir(exist_ok=True)
    with tempfile.TemporaryDirectory(dir=tmp_dir_base) as tmp_dir:
        result = instantiate_collection(
            ldb_dir,
            collection,
            tmp_dir,
            fmt,
        )

        if apply:
            paths = [str(ldb_dir / InstanceDir.USER_FILTERS)]
            apply_transform(apply, tmp_dir, os.fspath(dest), paths=paths)
        else:
            # check again to make sure nothing was added while writing to the
            # temporary location
            ensure_path_is_empty_workspace(dest, force)
            dest_str = os.fspath(dest)
            for path in Path(tmp_dir).iterdir():
                shutil.move(os.fspath(path), dest_str)

    return result


def instantiate_collection(
    ldb_dir: Path,
    collection: Mapping[str, Optional[str]],
    dest_dir: Union[str, Path],
    fmt: str,
) -> InstantiateResult:
    fmt = INSTANTIATE_FORMATS[fmt]
    if fmt in (Format.STRICT, Format.BARE):
        return copy_pairs(
            ldb_dir,
            collection,
            dest_dir,
            fmt == Format.STRICT,
        )
    if fmt == Format.ANNOT:
        return copy_annot(
            ldb_dir,
            collection,
            dest_dir,
        )
    if fmt == Format.INFER:
        return copy_infer(
            ldb_dir,
            collection,
            dest_dir,
        )
    if fmt == Format.LABEL_STUDIO:
        return copy_label_studio(
            ldb_dir,
            collection,
            dest_dir,
        )
    raise ValueError(f"Not a valid indexing format: {fmt}")


def apply_transform(
    proc_args: Sequence[str],
    input_dir: str,
    output_dir: str,
    paths: Sequence[str] = (),
) -> int:
    from ldb.pipe import open_plugin  # pylint: disable=import-outside-toplevel

    data = json.dumps(
        [
            os.path.abspath(input_dir),
            os.path.abspath(output_dir),
        ],
    )
    with open_plugin(proc_args, paths, set_cwd=True) as proc:
        stdout, stderr = proc.communicate(data)
        retcode = proc.poll() or 0
        if retcode:
            raise subprocess.CalledProcessError(
                retcode,
                proc.args,
                output=stdout,
                stderr=stderr,
            )
    return retcode


@dataclass
class InstItem:
    ldb_dir: Path
    dest_dir: Union[str, Path]
    data_object_hash: str
    storage_locations: Collection[StorageLocation]

    @cached_property
    def data_object_meta(self) -> JSONObject:
        data_object_dir = get_hash_path(
            self.ldb_dir / InstanceDir.DATA_OBJECT_INFO,
            self.data_object_hash,
        )
        return load_data_file(data_object_dir / "meta")  # type: ignore[no-any-return] # noqa: E501

    @cached_property
    def _prefix_ext(self) -> Tuple[str, str]:
        return get_prefix_ext(self.data_object_meta["fs"]["path"])

    @property
    def prefix(self) -> str:
        return self._prefix_ext[0]

    @property
    def ext(self) -> str:
        return self._prefix_ext[1]

    @cached_property
    def base_dest(self) -> str:
        return os.path.join(
            self.dest_dir,
            self.prefix + self.data_object_hash,
        )

    @cached_property
    def data_object_dest(self) -> str:
        return self.base_dest + self.ext.lower()

    def copy_data_object(
        self,
    ) -> str:
        fs_protocol: FSProtocol = self.data_object_meta["fs"]["protocol"]
        protocol: str = first_protocol(fs_protocol)
        path: str = self.data_object_meta["fs"]["path"]
        fs = get_filesystem(path, protocol, self.storage_locations)
        os.makedirs(os.path.split(self.data_object_dest)[0], exist_ok=True)
        dest = self.data_object_dest

        fs.get_file(path, dest)
        return dest


@dataclass
class PairInstItem(InstItem):
    annotation_hash: str

    @cached_property
    def annotation_content(self) -> JSONDecoded:
        return get_annotation(self.ldb_dir, self.annotation_hash)

    @cached_property
    def annotation_content_bytes(self) -> bytes:
        return serialize_annotation(self.annotation_content).encode()

    @cached_property
    def annotation_dest(self) -> str:
        return self.base_dest + ".json"

    def copy_annotation(self) -> str:
        dest = self.annotation_dest
        os.makedirs(os.path.split(dest)[0], exist_ok=True)
        write_data_file(
            Path(dest),
            self.annotation_content_bytes,
        )
        return dest


@dataclass
class AnnotationOnlyInstItem(PairInstItem):
    @cached_property
    def annotation_content(self) -> JSONDecoded:
        annotation: JSONObject = get_annotation(  # type: ignore[assignment]
            self.ldb_dir,
            self.annotation_hash,
        )
        annotation = {
            "ldb_meta": {"data_object_id": self.data_object_hash},
            "annotation": annotation,
        }
        return annotation


@dataclass
class InferInstItem(PairInstItem):
    annotation_hash: str

    @cached_property
    def base_dest(self) -> str:
        parts: List[str] = []
        try:
            label = self.annotation_content["label"]  # type: ignore[index, call-overload] # noqa: E501
        except Exception as exc:
            raise LDBException(
                "Annotations for tensorflow-inferred format should contain "
                '"label" key',
            ) from exc
        key: str
        while isinstance(label, dict):
            key, label = next(iter(label.items()))
            parts.append(key)
        parts.append(label)
        return os.path.join(
            self.dest_dir,
            *parts,
            self.prefix + self.data_object_hash,
        )


@dataclass
class LabelStudioInstItem(PairInstItem):
    _annotation_content: List[JSONObject]

    def __init__(
        self,
        ldb_dir: Path,
        dest_dir: Union[str, Path],
        annotation_content: List[JSONObject],
        storage_locations: Collection[StorageLocation],
    ):
        super().__init__(
            ldb_dir,
            dest_dir,
            "",
            storage_locations,
            "",
        )
        self._annotation_content = annotation_content

    @cached_property
    def annotation_content(self) -> JSONDecoded:
        return self._annotation_content

    @cached_property
    def base_dest(self) -> str:
        return os.path.join(
            self.dest_dir,
            "annotations",
        )


def serialize_annotation(annotation: JSONDecoded) -> str:
    return json_dumps(annotation, indent=2)


def get_prefix_ext(path: str) -> Tuple[str, str]:
    prefix = path.rsplit("/", 1)[-1].rsplit("\\", 1)[-1]
    prefix, ext = os.path.splitext(prefix)
    prefix = prefix.replace(".", "-") + "-"
    return prefix, ext


def copy_pairs(
    ldb_dir: Path,
    collection: Mapping[str, Optional[str]],
    dest_dir: Union[str, Path],
    strict: bool = False,
) -> InstantiateResult:
    items: List[PairInstItem] = []
    data_obj_paths: List[str] = []
    annot_paths = []
    num_annotations = 0
    # annotations are small and stored in ldb; copy them first
    for data_object_hash, annotation_hash in collection.items():
        if annotation_hash is None:
            if strict:
                continue
            annotation_hash = ""
        item = PairInstItem(
            ldb_dir,
            dest_dir,
            data_object_hash,
            get_storage_locations(ldb_dir),
            annotation_hash,
        )
        if annotation_hash:
            annot_paths.append(item.copy_annotation())
            num_annotations += 1
        else:
            annot_paths.append("")
        items.append(item)

    with ThreadPoolExecutor(max_workers=4 * (os.cpu_count() or 1)) as pool:
        with get_progressbar(transient=True) as progress:
            task = progress.add_task("Instantiate", total=len(items))

            def worker(item: PairInstItem) -> str:
                result = item.copy_data_object()
                progress.update(task, advance=1)
                return result

            data_obj_paths = list(pool.map(worker, items))

    return InstantiateResult(
        data_obj_paths,
        annot_paths,
        len(data_obj_paths),
        num_annotations,
    )


def copy_annot(
    ldb_dir: Path,
    collection: Mapping[str, Optional[str]],
    dest_dir: Union[str, Path],
) -> InstantiateResult:
    annot_paths = []
    for data_object_hash, annotation_hash in collection.items():
        if annotation_hash:
            path = AnnotationOnlyInstItem(
                ldb_dir,
                dest_dir,
                data_object_hash,
                get_storage_locations(ldb_dir),
                annotation_hash,
            ).copy_annotation()
            annot_paths.append(path)
    return InstantiateResult(
        [],
        annot_paths,
        0,
        len(annot_paths),
    )


def copy_infer(
    ldb_dir: Path,
    collection: Mapping[str, Optional[str]],
    dest_dir: Union[str, Path],
) -> InstantiateResult:
    for data_object_hash, annotation_hash in collection.items():
        if not annotation_hash:
            raise LDBException(
                "For tensorflow-inferred instantiate format, "
                "all data objects must have an annotation. "
                "Missing annotation for data object: "
                f"{DATA_OBJ_ID_PREFIX}{data_object_hash}",
            )
    data_obj_paths = []
    for data_object_hash, annotation_hash in cast(
        Mapping[str, str],
        collection,
    ).items():
        path = InferInstItem(
            ldb_dir,
            dest_dir,
            data_object_hash,
            get_storage_locations(ldb_dir),
            annotation_hash,
        ).copy_data_object()
        data_obj_paths.append(path)
    return InstantiateResult(
        data_obj_paths,
        [],
        len(data_obj_paths),
        0,
    )


def copy_label_studio(
    ldb_dir: Path,
    collection: Mapping[str, Optional[str]],
    dest_dir: Union[str, Path],
    url_key: str = "image",
) -> InstantiateResult:
    annot_paths = []
    annotations: List[JSONObject] = []
    for data_object_hash, annotation_hash in collection.items():
        if not annotation_hash:
            raise LDBException(
                "For label-studio instantiate format, "
                "all data objects must have an annotation. "
                "Missing annotation for data object: "
                f"{DATA_OBJ_ID_PREFIX}{data_object_hash}",
            )
        annot = get_annotation(ldb_dir, annotation_hash)
        try:
            annot["data"][url_key]  # type: ignore[call-overload, index]
        except Exception as exc:
            raise LDBException(
                "For label-studio instantiate format, "
                f'annotations must have the key "data.{url_key}." '
                "Malformatted annotation for data object: "
                f"{DATA_OBJ_ID_PREFIX}{data_object_hash}",
            ) from exc
        annotations.append(annot)  # type: ignore[arg-type]
    path = LabelStudioInstItem(
        ldb_dir,
        dest_dir,
        annotations,
        get_storage_locations(ldb_dir),
    ).copy_annotation()
    annot_paths.append(path)
    return InstantiateResult(
        [],
        annot_paths,
        0,
        len(annotations),
    )



from typing import Iterable, Iterator, TypeVar

T = TypeVar("T")


def take(iterable: Iterable[T], n: int = 1) -> Iterator[T]:
    for i, item in enumerate(iterable):
        if i < n:
            yield item



__all__ = ["main"]

import logging
import sys
from logging import Logger
from types import TracebackType
from typing import List, Optional, Tuple, Type, Union

from ldb.cli import get_main_parser

logger = logging.getLogger(__name__)

MIN_PYTHON_VERSION_INFO = (3, 7)
MIN_PYTHON_VERSION = ".".join(map(str, MIN_PYTHON_VERSION_INFO))


class QuietFormatter(logging.Formatter):
    def formatException(
        self,
        ei: Union[
            Tuple[Type[BaseException], BaseException, Optional[TracebackType]],
            Tuple[None, None, None],
        ],
    ) -> str:
        return ""


def configure_logger(log: Logger, verbose: bool) -> None:
    log = logging.getLogger("ldb")
    handler = logging.StreamHandler()
    fmt = "%(levelname)s: %(message)s"
    if verbose:
        formatter = logging.Formatter(fmt)
    else:
        formatter = QuietFormatter(fmt)
    handler.setFormatter(formatter)
    log.addHandler(handler)


def main(argv: Optional[List[str]] = None) -> int:
    main_parser = get_main_parser()
    options = main_parser.parse_args(args=argv)
    configure_logger(logger, options.verbose > 0)
    try:
        if sys.version_info < MIN_PYTHON_VERSION_INFO:
            raise ImportError(
                f"Python {MIN_PYTHON_VERSION} or greater is required. "
                f"Unsupported version running:\n{sys.version}",
            )
        try:
            func = options.func
        except AttributeError:
            main_parser.print_usage()
            return 1
        func(options)
    except Exception as exc:  # pylint: disable=broad-except
        logger.exception(exc)
        return 1
    return 0



from typing import Any, Iterable, Iterator, TypeVar

T = TypeVar("T")


def sort_by_iterable(
    data: Iterable[T],
    iterable: Iterable[Any],
) -> Iterator[T]:
    for _, d in sorted(
        zip(iterable, data),
        key=lambda x: x[0],  # type: ignore[no-any-return]
        reverse=True,
    ):
        yield d



import os
import shlex
import shutil
from pathlib import Path
from typing import Optional

from ldb import config
from ldb.config import get_default_instance_dir, get_global_base, get_ldb_dir
from ldb.exceptions import LDBException, LDBInstanceNotFoundError
from ldb.path import INSTANCE_DIRS, Filename, GlobalDir
from ldb.storage import StorageLocation, add_storage


def init(
    path: Path,
    force: bool = False,
    read_any_cloud_location: bool = False,
    auto_index: bool = False,
) -> Path:
    """
    Create a new LDB instance.
    """
    path = Path(os.path.normpath(path))
    if path.is_dir() and next(path.iterdir(), None) is not None:
        if is_ldb_instance(path):
            if force:
                print(
                    "Removing existing LDB instance at "
                    f"{repr(os.fspath(path))}",
                )
                shutil.rmtree(path)
            else:
                raise LDBException(
                    "Initialization failed\n"
                    "An LDB instance already exists at "
                    f"{repr(os.fspath(path))}\n"
                    "Use the --force option to remove it",
                )
        else:
            raise LDBException(
                f"Directory not empty: {repr(os.fspath(path))}\n"
                "To create an LDB instance here, remove directory contents",
            )
    for subdir in INSTANCE_DIRS:
        (path / subdir).mkdir(parents=True)
    with config.edit(path / Filename.CONFIG) as cfg:
        cfg["core"] = {
            "read_any_cloud_location": read_any_cloud_location,
            "auto_index": auto_index,
        }
    print(f"Initialized LDB instance at {repr(os.fspath(path))}")
    return path


def init_quickstart(force: bool = False) -> Path:
    ldb_dir = init(
        get_default_instance_dir(),
        force=force,
        read_any_cloud_location=True,
        auto_index=True,
    )
    add_default_read_add_storage(ldb_dir)
    return ldb_dir


def add_default_read_add_storage(ldb_dir: Path) -> None:
    path = get_global_base() / GlobalDir.DEFAULT_READ_ADD_STORAGE
    try:
        path.mkdir(parents=True, exist_ok=True)
    except OSError as exc:
        path_arg = shlex.quote(os.fspath(path))
        raise LDBException(
            f"Unable to create read-add storage location: {path_arg}\n"
            "Ensure it is a writable directory and add it with:\n\n"
            f"  ldb add-storage -a {path_arg}\n",
        ) from exc
    add_storage(
        ldb_dir / Filename.STORAGE,
        StorageLocation(
            path=os.fspath(path),
            read_and_add=True,
        ),
    )


def is_ldb_instance(path: Path) -> bool:
    return all((path / subdir).is_dir() for subdir in INSTANCE_DIRS)


def get_ldb_instance(path: Optional[Path] = None) -> Path:
    if path is None:
        path = get_ldb_dir()
    if not is_ldb_instance(path):
        raise LDBInstanceNotFoundError(
            f"No LDB instance at {os.fspath(path)!r}",
        )
    return path



from argparse import ArgumentParser, Namespace
from typing import TYPE_CHECKING, Iterable

import shtab

from ldb import config
from ldb.cli_utils import add_data_format_arguments, tag_list
from ldb.config import ConfigType
from ldb.core import get_ldb_instance
from ldb.data_formats import INDEX_FORMATS, Format
from ldb.index import index

if TYPE_CHECKING:
    from argparse import _SubParsersAction


def index_command(options: Namespace) -> None:
    tags = [t for lst in options.tags for t in lst]
    ldb_dir = get_ldb_instance()
    result = index(
        ldb_dir,
        options.paths,
        read_any_cloud_location=(
            (
                config.load_first([ConfigType.INSTANCE])  # type: ignore[call-overload] # noqa: E501
                or {}
            )
            .get("core", {})
            .get("read_any_cloud_location", False)
        ),
        fmt=options.format,
        tags=tags,
    )
    print(result.summary())


def add_parser(
    subparsers: "_SubParsersAction[ArgumentParser]",
    parents: Iterable[ArgumentParser],
) -> None:
    parser = subparsers.add_parser(
        "index",
        parents=parents,
        help="Index a storage location",
    )
    add_data_format_arguments(
        parser,
        default=Format.AUTO,
        formats=INDEX_FORMATS,
    )
    parser.add_argument(
        "--add-tags",
        metavar="<tags>",
        dest="tags",
        default=[],
        type=tag_list,
        action="append",
        help="Comma-separated list of tags to add to indexed data objects",
    )
    parser.add_argument(  # type: ignore[attr-defined]
        "paths",
        metavar="<path>",
        nargs="+",
        help="File, directory, or path prefix. Supports glob syntax.",
    ).complete = shtab.FILE
    parser.set_defaults(func=index_command)



from typing import (
    Any,
    Callable,
    Dict,
    List,
    Mapping,
    MutableMapping,
    Sequence,
    Tuple,
    Union,
)

from jmespath.visitor import _Expression as JMESPathExpression
from typing_extensions import Protocol

JSONKey = Union[str, int, float, bool, None]
JSONObject = Dict[JSONKey, Any]
JSONArray = List[Any]
# possible return values of `json.decoder.JSONDecoder`
# `Any` should be `JSONDecoded`, but mypy doesn't support cyclic definitions
JSONDecoded = Union[str, int, float, bool, None, JSONObject, JSONArray]
JMESPathValue = Union[JSONDecoded, JMESPathExpression]


class JSONFunc(Protocol):
    def __call__(self, *args: JMESPathValue) -> JMESPathValue:
        ...


class JSONInstanceFunc(Protocol):
    def __call__(self, self_: Any, *args: JMESPathValue) -> JMESPathValue:
        ...


JSONArgTypes = Sequence[str]
JSONFuncDef = Tuple[JSONFunc, JSONArgTypes]
JSONFuncMapping = Mapping[str, JSONFuncDef]
JSONFuncMutableMapping = MutableMapping[str, JSONFuncDef]
JSONBinFunc = Callable[[JMESPathValue, JMESPathValue], JSONDecoded]



class OpType:
    ANNOTATION_QUERY = "annotation_query"
    JP_ANNOTATION_QUERY = "jp_annotation_query"
    FILE_QUERY = "file_query"
    PATH_QUERY = "path_query"
    TAG_QUERY = "tag_query"
    NO_TAG_QUERY = "no_tag_query"
    LIMIT = "limit"
    SAMPLE = "sample"
    PIPE = "pipe"



import json
import re
from argparse import ArgumentParser, Namespace
from typing import TYPE_CHECKING, Iterable, Union

from ldb.cli_utils import add_data_obj_params
from ldb.core import get_ldb_instance
from ldb.evaluate import evaluate
from ldb.utils import DATA_OBJ_ID_PREFIX

if TYPE_CHECKING:
    from argparse import _SubParsersAction


def evaluate_command(options: Namespace) -> None:
    indent = get_indent_value(options.indent)
    for data_object_hash, *results in evaluate(
        get_ldb_instance(),
        options.paths,
        options.query_args,
        warn=True,
    ):
        if not options.json_only:
            print(f"{DATA_OBJ_ID_PREFIX}{data_object_hash}")
        for item in results:
            print(json.dumps(item, indent=indent))
        if not options.json_only:
            print()


def get_indent_value(indent: str) -> Union[str, int, None]:
    if indent == "none":
        return None
    if re.search(r"^\s*$", indent):
        return indent
    if re.search(r"^\d+$", indent):
        return int(indent)
    raise ValueError(f"Invalid indent string: {indent!r}")


def add_parser(
    subparsers: "_SubParsersAction[ArgumentParser]",
    parents: Iterable[ArgumentParser],
) -> None:
    parser = subparsers.add_parser(
        "eval",
        parents=parents,
        help="Evaluate a query on the specified annotations",
    )
    parser.add_argument(
        "-j",
        "--json-only",
        action="store_true",
        default=False,
        help="Show JSON output only instead of showing object hashes",
    )
    parser.add_argument(
        "--indent",
        metavar="<value>",
        default="2",
        type=str,
        help=(
            "Indentation for JSON output as a whitespace string or an "
            "integer specifying the number of spaces"
        ),
    )
    add_data_obj_params(parser, dest="query_args")
    parser.set_defaults(func=evaluate_command)



from dataclasses import dataclass
from pathlib import Path
from typing import Iterable, List, Optional, Sequence, Tuple

from ldb.add import process_args_for_ls
from ldb.dataset import OpDef, apply_queries
from ldb.path import InstanceDir
from ldb.string_utils import left_truncate
from ldb.utils import DATA_OBJ_ID_PREFIX, get_hash_path, load_data_file


@dataclass
class DatasetListing:
    data_object_hash: str
    data_object_path: str
    annotation_hash: str
    annotation_version: int


def ls(
    ldb_dir: Path,
    paths: Sequence[str],
    collection_ops: Iterable[OpDef],
    warn: bool = True,
) -> List[DatasetListing]:
    data_object_hashes, annotation_hashes, _ = process_args_for_ls(
        ldb_dir,
        paths,
    )
    collection = apply_queries(
        ldb_dir,
        data_object_hashes,
        annotation_hashes,
        collection_ops,
        warn=warn,
    )
    return ls_collection(ldb_dir, collection)


def print_dataset_listings(
    dataset_listings: List[DatasetListing],
    verbose: bool = False,
) -> int:
    if not dataset_listings:
        return 0

    num_items = 0
    print(f"{'Data Object Hash':37} {'Annot.':8} Data Object Path")
    for item in dataset_listings:
        annotation_version = str(item.annotation_version or "-")
        path = (
            item.data_object_path
            if verbose
            else left_truncate(item.data_object_path)
        )
        print(
            f"  {DATA_OBJ_ID_PREFIX}{item.data_object_hash:35} "
            f"{annotation_version:8} {path}",
        )
        num_items += 1
    return num_items


def ls_collection(
    ldb_dir: Path,
    collection: Iterable[Tuple[str, Optional[str]]],
) -> List[DatasetListing]:
    result = []
    data_object_info_path = ldb_dir / InstanceDir.DATA_OBJECT_INFO
    for data_object_hash, annotation_hash in collection:
        data_object_dir = get_hash_path(
            data_object_info_path,
            data_object_hash,
        )
        annotation_version = 0
        if annotation_hash:
            annotation_meta = load_data_file(
                data_object_dir / "annotations" / annotation_hash,
            )
            annotation_version = annotation_meta["version"]
        data_object_meta = load_data_file(data_object_dir / "meta")

        result.append(
            DatasetListing(
                data_object_hash=data_object_hash,
                data_object_path=data_object_meta["fs"]["path"],
                annotation_hash=annotation_hash or "",
                annotation_version=annotation_version,
            ),
        )
    return result



from dataclasses import dataclass
from functools import wraps
from typing import Any, Dict, List, Optional, Sequence, Type, Union

from jmespath import functions
from jmespath.visitor import Options

from ldb.config import get_ldb_dir
from ldb.core import is_ldb_instance
from ldb.query.functions import CUSTOM_FUNCTIONS
from ldb.query.user_functions import load_user_functions
from ldb.typing import (
    JMESPathValue,
    JSONArgTypes,
    JSONFunc,
    JSONFuncMapping,
    JSONFuncMutableMapping,
    JSONInstanceFunc,
)

VALID_ARG_TYPES = set(functions.REVERSE_TYPES_MAP)


def create_func(
    func: JSONFunc,
    arg_types: JSONArgTypes,
) -> JSONInstanceFunc:
    variadic = False
    if arg_types and arg_types[-1] == "*":
        variadic = True
        arg_types = arg_types[:-1]

    signature: List[Dict[str, Union[Sequence[str], bool]]] = []
    for arg in arg_types:
        types = arg.split("|")
        for type_str in types:
            validate_arg_type(type_str)
        signature.append({"types": types})

    if signature and variadic:
        signature[-1]["variadic"] = True

    @wraps(func)
    @functions.signature(*signature)
    def new_func(_self: Any, *args: JMESPathValue) -> JMESPathValue:
        return func(*args)

    return new_func


def validate_arg_type(arg_type: str) -> None:
    types = arg_type.split("-", 1)
    if len(types) == 2:
        if types[0] != "array":
            raise ValueError(
                f"Invalid arg type: {arg_type!r}\n"
                f"{types[0]!r} does not support sub-types. "
                "Only 'array' does.",
            )
        if types[1] not in VALID_ARG_TYPES:
            raise ValueError(
                f"Invalid arg type: {arg_type!r}\n"
                f"{types[1]!r} is not a valid sub-type.\n"
                f"Use one of: {VALID_ARG_TYPES}",
            )
    elif types[0] not in VALID_ARG_TYPES:
        raise ValueError(
            f"Invalid arg type: {arg_type!r}\n"
            f"Use one of: {VALID_ARG_TYPES}",
        )


def create_func_holder(custom_functions: JSONFuncMapping) -> Type[Any]:
    class FuncHolder:
        pass

    for name, (func, arg_types) in custom_functions.items():
        setattr(FuncHolder, "_func_" + name, create_func(func, arg_types))
    return FuncHolder


def create_custom_options(custom_functions: JSONFuncMapping) -> Options:
    FuncHolder = create_func_holder(custom_functions)

    class CustomFunctions(
        functions.Functions,
        FuncHolder,  # type: ignore[misc,valid-type]
    ):
        pass

    return Options(custom_functions=CustomFunctions())


def get_all_custom_functions() -> JSONFuncMutableMapping:
    custom_functions: JSONFuncMutableMapping = (
        CUSTOM_FUNCTIONS.copy()  # type: ignore[assignment]
    )
    ldb_dir = get_ldb_dir()
    if is_ldb_instance(ldb_dir):
        custom_functions.update(load_user_functions(ldb_dir, CUSTOM_FUNCTIONS))
    return custom_functions


@dataclass(init=False)
class OptionsCache:
    """
    Cache for lazy loading of options for JMESPath queries.

    User-defined custom functions may be any python file and importing
    could be expensive. A new instance of this class will not load user-
    defined custom functions until the first call to the `get` method.
    Then subsequent calls to `get` return the cached `Options` object.
    """

    _options: Optional[Options] = None

    def get(self) -> Options:
        if self._options is None:
            self._options = create_custom_options(get_all_custom_functions())
        return self._options



from typing import Optional, Sequence, TypeVar

import torch
from PIL import Image
from torch.nn import AdaptiveAvgPool2d, CosineSimilarity, Module
from torchvision import models, transforms
from torchvision.models.feature_extraction import create_feature_extractor
from torchvision.transforms import InterpolationMode

T = TypeVar("T", bound="ResNetLayer")

MODELS = {
    "18": models.resnet18,
    "34": models.resnet34,
    "50": models.resnet50,
    "101": models.resnet101,
    "152": models.resnet152,
}
BLOCK_EXPANSION = {
    "18": 1,
    "34": 1,
    "50": 4,
    "101": 4,
    "152": 4,
}
RETURN_NODES = {
    "layer1": "layer1",
    "layer2": "layer2",
    "layer3": "layer3",
    "layer4": "layer4",
}
LAYER_NAMES = ("layer1", "layer2", "layer3", "layer4")

preprocess = transforms.Compose(
    [
        transforms.Resize(256, interpolation=InterpolationMode.BICUBIC),
        transforms.CenterCrop(224),
        (lambda x: x.convert("RGB")),
        transforms.ToTensor(),
        transforms.Normalize(
            mean=[0.485, 0.456, 0.406],
            std=[0.229, 0.224, 0.225],
        ),
    ],
)
cos_sim = CosineSimilarity()


class ResNetLayer(Module):
    def __init__(self, model_name: str, layer_index: int) -> None:
        super().__init__()
        self.model_name = model_name
        self.layer_index = layer_index
        self.model = MODELS[self.model_name](pretrained=True)
        self.body = create_feature_extractor(
            self.model,
            return_nodes=RETURN_NODES,
        )
        self.avgpool = AdaptiveAvgPool2d((1, 1))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        features = self.body(x)
        x = self.avgpool(features[LAYER_NAMES[self.layer_index]])
        x = torch.flatten(x, 1)
        return x

    def eval(self: T) -> T:
        self.model.eval()
        return self


def image_layer_similarity(
    image_file_path: str,
    file_paths: Sequence[str],
    model_name: str = "18",
    layer_index: Optional[int] = None,
) -> torch.Tensor:
    if layer_index is None:
        model = MODELS[model_name](pretrained=True)
    else:
        model = ResNetLayer(model_name, layer_index)
    model.eval()

    all_file_paths = [image_file_path, *file_paths]
    input_tensors = [preprocess(Image.open(x)) for x in all_file_paths]
    input_batch = torch.stack(input_tensors)

    if torch.cuda.is_available():
        input_batch = input_batch.to("cuda")
        model.to("cuda")

    with torch.no_grad():  # type: ignore[no-untyped-call]
        output = model(input_batch)

    return cos_sim(output[:1], output[1:])  # type: ignore[no-any-return]



import os
from pathlib import Path
from typing import Iterable, Iterator, List, Sequence, Tuple

from ldb.add import process_args_for_ls
from ldb.dataset import (
    OpDef,
    apply_queries_to_collection,
    get_collection_dir_keys,
)
from ldb.path import InstanceDir, WorkspacePath
from ldb.typing import JSONObject
from ldb.utils import format_dataset_identifier, get_hash_path, load_data_file
from ldb.workspace import load_workspace_dataset


def pull(
    ldb_dir: Path,
    workspace_path: Path,
    paths: Sequence[str],
    collection_ops: Iterable[OpDef],
    version: int,
    warn: bool = True,
) -> None:
    version_msg = "latest version" if version == -1 else f"v{version}"
    ds_name = load_workspace_dataset(workspace_path).dataset_name
    ds_ident = format_dataset_identifier(ds_name)
    ws_data_object_hashes = set(
        get_collection_dir_keys(workspace_path / WorkspacePath.COLLECTION),
    )
    data_object_hashes, annotation_hashes, _ = process_args_for_ls(
        ldb_dir,
        paths,
    )
    collection: Iterator[Tuple[str, str]] = (
        (d, a)
        for d, a in zip(data_object_hashes, annotation_hashes)
        if d in ws_data_object_hashes
    )
    collection = apply_queries_to_collection(
        ldb_dir,
        collection,
        collection_ops,
        warn=warn,
    )
    data_object_hashes = (d for d, _ in collection)
    print(f"Updating to {version_msg}")
    updates, num_missing = get_collection_with_updated_annotations(
        ldb_dir,
        data_object_hashes,
        version,
    )
    num_updated_annots, num_already_updated = update_annotation_versions(
        workspace_path,
        updates,
    )
    print(
        "\n"
        f"{ds_ident}\n"
        f"  Already up-to-date: {num_already_updated:8d}\n"
        f"  New updates:        {num_updated_annots:8d}\n"
        f"  Missing version:    {num_missing:8d}\n",
    )


def get_hash_annot_pair_version(
    hash_annot_pair: Tuple[str, JSONObject],
) -> int:
    return hash_annot_pair[1]["version"]  # type: ignore[no-any-return]


def get_annotation_version_hash(
    ldb_dir: Path,
    data_object_hash: str,
    version: int,
) -> str:
    data_object_dir = get_hash_path(
        ldb_dir / InstanceDir.DATA_OBJECT_INFO,
        data_object_hash,
    )
    if (data_object_dir / "annotations").is_dir():
        files = list((data_object_dir / "annotations").iterdir())
        if version == -1:
            version = len(files)
        for file in files:
            annot = load_data_file(file)
            if annot["version"] == version:
                return file.name
    return ""


def get_collection_with_updated_annotations(
    ldb_dir: Path,
    data_object_hashes: Iterable[str],
    version: int = -1,
) -> Tuple[List[Tuple[str, str]], int]:
    result = []
    num_missing = 0
    for data_object_hash in data_object_hashes:
        new_annotation_hash = get_annotation_version_hash(
            ldb_dir,
            data_object_hash,
            version,
        )
        if new_annotation_hash:
            result.append((data_object_hash, new_annotation_hash))
        else:
            num_missing += 1
    return result, num_missing


def update_annotation_versions(
    workspace_path: Path,
    collection: Iterable[Tuple[str, str]],
) -> Tuple[int, int]:
    workspace_path = Path(os.path.normpath(workspace_path))
    collection_dir_path = workspace_path / WorkspacePath.COLLECTION

    to_write = []
    for data_object_hash, annotation_hash in collection:
        to_write.append(
            (
                get_hash_path(collection_dir_path, data_object_hash),
                annotation_hash,
            ),
        )

    num_updated_annots = 0
    num_already_updated = 0
    for collection_member_path, annotation_hash in to_write:
        with collection_member_path.open("r+") as file:
            existing_annotation_hash = file.read()
            if annotation_hash == existing_annotation_hash:
                num_already_updated += 1
            else:
                file.seek(0)
                file.write(annotation_hash)
                file.truncate()
                num_updated_annots += 1
    return num_updated_annots, num_already_updated



from pathlib import Path
from typing import Collection, Iterable, Tuple

from ldb.path import InstanceDir
from ldb.utils import (
    get_hash_path,
    json_dumps,
    load_data_file,
    write_data_file,
)


def tag_data_objects(
    ldb_dir: Path,
    data_object_hashes: Iterable[str],
    add_tags: Collection[str] = (),
    remove_tags: Collection[str] = (),
) -> Tuple[int, int]:
    print("Tagging data objects")
    num_selected = 0
    num_changed = 0
    for data_object_hash in data_object_hashes:
        num_changed += tag_data_object(
            ldb_dir,
            data_object_hash,
            add_tags,
            remove_tags,
        )
        num_selected += 1
    print(
        f"  Data objects: {num_selected:8d}\n"
        f"  Num updated:  {num_changed:8d}",
    )
    return num_selected, num_changed


def tag_data_object(
    ldb_dir: Path,
    data_object_hash: str,
    add_tags: Collection[str] = (),
    remove_tags: Collection[str] = (),
) -> bool:
    meta_path = (
        get_hash_path(
            ldb_dir / InstanceDir.DATA_OBJECT_INFO,
            data_object_hash,
        )
        / "meta"
    )
    meta = load_data_file(meta_path)
    tag_set = set(meta["tags"])
    tag_set.update(add_tags)
    tag_set.difference_update(remove_tags)
    result = sorted(tag_set)
    if result != meta["tags"]:
        meta["tags"] = result
        write_data_file(meta_path, json_dumps(meta).encode())
        return True
    return False



class LDBException(Exception):
    pass


class LDBInstanceNotFoundError(LDBException):
    pass


class DatasetError(LDBException):
    pass


class DatasetNotFoundError(DatasetError):
    pass


class DataObjectNotFoundError(LDBException):
    pass


class WorkspaceError(LDBException):
    pass


class WorkspaceDatasetNotFoundError(WorkspaceError):
    pass


class StorageConfigurationError(LDBException):
    pass


class NotAStorageLocationError(LDBException):
    pass


class IndexingException(LDBException):
    pass



import re
import warnings
from typing import Callable, Collection, Iterable, Iterator

import jmespath
from jmespath.exceptions import JMESPathError, ParseError

from ldb.jmespath import custom_compile, jp_compile
from ldb.query.utils import OptionsCache
from ldb.typing import JSONDecoded
from ldb.warnings import SimpleWarningHandler

SearchFunc = Callable[[Iterable[JSONDecoded]], Iterator[JSONDecoded]]
BoolSearchFunc = Callable[[Iterable[JSONDecoded]], Iterator[bool]]

OPTIONS_CACHE = OptionsCache()


def get_search_func(
    query_str: str,
    use_custom: bool = True,
    warn: bool = True,
) -> SearchFunc:
    """
    Compile `query_str` and return a search function.
    """
    if use_custom:
        query_obj = custom_compile(query_str)
    else:
        query_obj = jp_compile(query_str)

    runtime_exc_types = set()

    def search(objects: Iterable[JSONDecoded]) -> Iterator[JSONDecoded]:
        with warnings.catch_warnings():
            warnings.showwarning = SimpleWarningHandler.showwarning
            for obj in objects:
                try:
                    yield query_obj.search(obj, options=OPTIONS_CACHE.get())
                except JMESPathError as exc:
                    if warn:
                        exc_type = type(exc)
                        if exc_type not in runtime_exc_types:
                            warnings.warn(
                                f"{exc_type.__name__}: {exc}",
                                RuntimeWarning,
                                stacklevel=2,
                            )
                        runtime_exc_types.add(exc_type)
                    yield None

    return search


def get_bool_search_func(
    query_str: str,
    use_custom: bool = True,
    warn: bool = True,
) -> BoolSearchFunc:
    """
    Adapt `query_str` to cast result to a bool and return a search function.

    This uses JMESPath's boolean rules, not Python's, so the following
    evaluate to false:

        Empty list: []
        Empty object: {}
        Empty string: ""
        False boolean: false
        Null value: null

    See https://jmespath.org/specification.html#or-expressions
    """
    modified_query_str = f"{query_str} && `true` || `false`"
    try:
        if use_custom:
            query_obj = custom_compile(modified_query_str)
        else:
            query_obj = jp_compile(modified_query_str)
    except ParseError:
        # If the expression above raises a ParseError, this should too
        # This way we show the original query string in the error message
        jmespath.compile(query_str)
        raise

    runtime_exc_types = set()

    def search(objects: Iterable[JSONDecoded]) -> Iterator[bool]:
        with warnings.catch_warnings():
            warnings.showwarning = SimpleWarningHandler.showwarning
            for obj in objects:
                try:
                    yield query_obj.search(  # type: ignore[misc]
                        obj,
                        options=OPTIONS_CACHE.get(),
                    )
                except JMESPathError as exc:
                    if warn:
                        exc_type = type(exc)
                        if exc_type not in runtime_exc_types:
                            warnings.warn(
                                f"{exc_type.__name__}: {exc}",
                                RuntimeWarning,
                                stacklevel=2,
                            )
                        runtime_exc_types.add(exc_type)
                    yield False

    return search


def get_tag_func(tags: Collection[str]) -> BoolSearchFunc:
    def search(objects: Iterable[Collection[str]]) -> Iterator[bool]:
        for obj in objects:
            yield any(t in obj for t in tags)

    return search  # type: ignore[return-value]


def get_no_tag_func(tags: Collection[str]) -> BoolSearchFunc:
    def search(objects: Iterable[Collection[str]]) -> Iterator[bool]:
        for obj in objects:
            yield any(t not in obj for t in tags)

    return search  # type: ignore[return-value]


def get_path_func(pattern: str) -> BoolSearchFunc:
    re_obj = re.compile(pattern)

    def search(objects: Iterable[Collection[str]]) -> Iterator[bool]:
        for obj in objects:
            for path in obj:
                if re_obj.search(path) is not None:
                    yield True
                    break
            else:
                yield False

    return search  # type: ignore[return-value]



import json
import os
import shutil
import subprocess
from contextlib import contextmanager
from pathlib import Path
from subprocess import Popen
from tempfile import TemporaryDirectory
from typing import (
    IO,
    TYPE_CHECKING,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
)

from ldb.data_formats import Format
from ldb.exceptions import DataObjectNotFoundError
from ldb.instantiate import instantiate_collection
from ldb.path import InstanceDir

if TYPE_CHECKING:
    StrPopen = Popen[str]  # pylint: disable=unsubscriptable-object)
else:
    StrPopen = Popen


def sort_collection(
    ldb_dir: Path,
    collection: Iterable[Tuple[str, str]],
    proc_args: List[str],
) -> Iterator[Tuple[str, str]]:
    collection_dict = dict(collection)
    with TemporaryDirectory() as temp_dir:
        result = instantiate_collection(
            ldb_dir,
            collection_dict,
            temp_dir,
            Format.BARE,
        )
        data_sequences = (
            collection_dict,
            result.data_object_paths,
            result.annotation_paths,
        )
        data = json.dumps(list(zip(*data_sequences)))
        paths = [str(ldb_dir / InstanceDir.USER_FILTERS)]
        for data_obj_hash in run_sort_process(proc_args, data, paths=paths):
            try:
                annot_hash = collection_dict[data_obj_hash]
            except KeyError:
                raise DataObjectNotFoundError(data_obj_hash) from KeyError
            yield data_obj_hash, annot_hash


@contextmanager
def open_proc(
    proc_args: Sequence[str],
    cwd: Optional[str] = None,
) -> Iterator[StrPopen]:
    with Popen(
        proc_args,
        stdout=subprocess.PIPE,
        stdin=subprocess.PIPE,
        text=True,
        cwd=cwd,
    ) as proc:
        yield proc


@contextmanager
def open_plugin(
    proc_args: Sequence[str],
    paths: Sequence[str] = (),
    set_cwd: bool = False,
) -> Iterator[StrPopen]:
    proc_path, other_args = proc_args[0], proc_args[1:]
    cwd = None
    if paths and not os.path.split(proc_path)[0]:
        for path in paths:
            full_path = shutil.which(proc_path, path=path)
            if full_path is not None:
                try:
                    if set_cwd:
                        cwd = os.path.split(full_path)[0] or None
                    with open_proc([full_path, *other_args], cwd=cwd) as proc:
                        yield proc
                except OSError:
                    pass
                else:
                    return
    if set_cwd:
        full_path = shutil.which(proc_path)
        if full_path is not None:
            cwd = os.path.split(full_path)[0] or None
            proc_path = full_path
    with open_proc([proc_path, *other_args], cwd=cwd) as proc:
        yield proc


def run_sort_process(
    proc_args: Sequence[str],
    data: str,
    paths: Sequence[str] = (),
) -> Iterator[str]:
    with open_plugin(proc_args, paths) as proc:
        stdin: IO[str] = proc.stdin  # type: ignore[assignment]
        stdout: IO[str] = proc.stdout  # type: ignore[assignment]
        stdin.write(data)
        stdin.close()
        for line in stdout.readlines():
            yield line.rstrip()



import sys
import warnings
from typing import Callable, Optional, TextIO, Type, Union
from warnings import WarningMessage

FormatFunc = Callable[[WarningMessage], str]


class WarningHandler:
    _formatwarnmsg_impl: FormatFunc = warnings._formatwarnmsg_impl  # type: ignore[attr-defined] # pylint: disable=protected-access # noqa: E501

    @classmethod
    def showwarning(
        cls,
        message: Union[Warning, str],
        category: Type[Warning],
        filename: str,
        lineno: int,
        file: Optional[TextIO] = None,
        line: Optional[str] = None,
    ) -> None:
        """
        Write a warning to a file, based on warnings.showwarning.
        """
        msg = WarningMessage(message, category, filename, lineno, file, line)
        cls._showwarnmsg_impl(msg)

    @classmethod
    def formatwarning(
        cls,
        message: Union[Warning, str],
        category: Type[Warning],
        filename: str,
        lineno: int,
        line: Optional[str] = None,
    ) -> str:
        """
        Format a warning, based on warnings.formatwarning.
        """
        msg = WarningMessage(message, category, filename, lineno, None, line)
        return cls._formatwarnmsg_impl(msg)

    @classmethod
    def _showwarnmsg_impl(cls, msg: WarningMessage) -> None:
        """
        Print a WarningMessage object to a file.

        Based on warnings._showwarnmsg_impl.
        """
        file = msg.file
        if file is None:
            file = sys.stderr
            if file is None:
                # sys.stderr is None when run with pythonw.exe:
                # warnings get lost
                return  # type: ignore[unreachable]
        text = cls._formatwarnmsg_impl(msg)
        try:
            file.write(text)
        except OSError:
            # the file (probably stderr) is invalid - this warning gets lost.
            pass


def simple_formatwarnmsg(msg: WarningMessage) -> str:
    return f"{msg.category.__name__}: {msg.message}\n"


class SimpleWarningHandler(WarningHandler):
    _formatwarnmsg_impl = simple_formatwarnmsg



import os
import shutil
from dataclasses import asdict, dataclass
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, Iterator, List, Optional, Union

from ldb.dataset import get_collection, get_collection_dir_items
from ldb.exceptions import WorkspaceDatasetNotFoundError, WorkspaceError
from ldb.path import WorkspacePath
from ldb.utils import (
    DATASET_PREFIX,
    ROOT,
    format_datetime,
    load_data_file,
    parse_datetime,
)


@dataclass
class WorkspaceDataset:
    dataset_name: str
    staged_time: datetime
    parent: str
    tags: List[str]

    @classmethod
    def parse(cls, attr_dict: Dict[str, Any]) -> "WorkspaceDataset":
        attr_dict = attr_dict.copy()
        return cls(
            staged_time=parse_datetime(attr_dict.pop("staged_time")),
            tags=attr_dict.pop("tags").copy(),
            **attr_dict,
        )

    def format(self) -> Dict[str, Any]:
        attr_dict = asdict(self)
        return dict(
            staged_time=format_datetime(attr_dict.pop("staged_time")),
            tags=attr_dict.pop("tags").copy(),
            **attr_dict,
        )


def workspace_dataset_is_clean(
    ldb_dir: Path,
    workspace_dataset_obj: WorkspaceDataset,
    workspace_path: Path,
) -> bool:
    ws_collection = collection_dir_to_object(
        workspace_path / WorkspacePath.COLLECTION,
    )
    if not workspace_dataset_obj.parent:
        return not ws_collection
    collection_obj = get_collection(ldb_dir, workspace_dataset_obj.parent)
    return ws_collection == collection_obj


def load_workspace_dataset(workspace_path: Path) -> WorkspaceDataset:
    try:
        workspace_ds = WorkspaceDataset.parse(
            load_data_file(workspace_path / WorkspacePath.DATASET),
        )
    except FileNotFoundError as exc:
        raise WorkspaceDatasetNotFoundError(
            "No workspace dataset staged at "
            f"{repr(os.fspath(workspace_path))}",
        ) from exc
    if workspace_ds.dataset_name == ROOT:
        raise ValueError(
            "Invalid workspace dataset name: "
            f"{DATASET_PREFIX}{workspace_ds.dataset_name}",
        )
    return workspace_ds


def collection_dir_to_object(collection_dir: Path) -> Dict[str, Optional[str]]:
    return dict(
        sorted(get_collection_dir_items(collection_dir, is_workspace=True)),
    )


def iter_workspace_dir(
    workspace_path: Union[str, Path],
) -> Iterator[os.DirEntry]:  # type: ignore[type-arg]
    ldb_workspace_name = WorkspacePath.BASE.name
    for entry in os.scandir(workspace_path):
        if entry.name != ldb_workspace_name or not entry.is_dir():
            yield entry


def ensure_path_is_empty_workspace(
    path: Union[str, Path],
    force: bool = False,
) -> None:
    if any(iter_workspace_dir(path)):
        try:
            load_workspace_dataset(Path(path))
        except WorkspaceDatasetNotFoundError as exc:
            raise WorkspaceDatasetNotFoundError(
                f"Not a workspace or an empty directory: {os.fspath(path)}",
            ) from exc
        if force:
            remove_workspace_contents(path)
        else:
            raise WorkspaceError(
                "Workspace is not empty: "
                f"{os.fspath(path)!r}\n"
                "Use the --force option to delete workspace contents",
            )


def remove_workspace_contents(workspace_path: Union[str, Path]) -> None:
    for entry in iter_workspace_dir(workspace_path):
        if entry.is_dir():
            shutil.rmtree(entry)
        else:
            os.remove(entry)



"""
Originally from dvc.fs.path.

Copied version:
  https://github.com/iterative/dvc/blob/e80e725fe45f74141e273dccf47f6a61b942b412/dvc/fs/path.py
Current version:
  https://github.com/iterative/dvc/blob/main/dvc/fs/path.py
"""
import ntpath
import posixpath
from typing import List, Tuple


class Path:
    def __init__(self, sep: str) -> None:
        if sep == posixpath.sep:
            self.flavour = posixpath
        elif sep == ntpath.sep:
            self.flavour = ntpath  # type: ignore[misc]
        else:
            raise ValueError(f"unsupported separator '{sep}'")

    def join(self, *parts: str) -> str:
        return self.flavour.join(*parts)

    def parts(self, path: str) -> Tuple[str, ...]:
        drive, path = self.flavour.splitdrive(path)

        ret: List[str] = []
        while True:
            path, part = self.flavour.split(path)

            if part:
                ret.append(part)
                continue

            if path:
                ret.append(path)

            break

        ret.reverse()

        if drive:
            ret = [drive] + ret

        return tuple(ret)

    def parent(self, path: str) -> str:
        return self.flavour.dirname(path)

    def parents(self, path: str) -> Tuple[str, ...]:
        parts = self.parts(path)
        return tuple(
            self.join(*parts[:length])
            for length in range(len(parts) - 1, 0, -1)
        )

    def name(self, path: str) -> str:
        return self.parts(path)[-1]

    def suffix(self, path: str) -> str:
        name = self.name(path)
        _, dot, suffix = name.partition(".")
        return dot + suffix

    def with_name(self, path: str, name: str) -> str:
        parts = list(self.parts(path))
        parts[-1] = name
        return self.join(*parts)

    def with_suffix(self, path: str, suffix: str) -> str:
        parts = list(self.parts(path))
        real_path, _, _ = parts[-1].partition(".")
        parts[-1] = real_path + suffix
        return self.join(*parts)

    def isin(self, left: str, right: str) -> bool:
        left_parts = self.parts(left)
        right_parts = self.parts(right)
        left_len = len(left_parts)
        right_len = len(right_parts)
        return left_len > right_len and left_parts[:right_len] == right_parts

    def isin_or_eq(self, left: str, right: str) -> bool:
        return left == right or self.isin(left, right)

    def overlaps(self, left: str, right: str) -> bool:
        # pylint: disable=arguments-out-of-order
        return self.isin_or_eq(left, right) or self.isin(right, left)

    def relpath(self, path: str, start: str) -> str:
        assert start
        return self.flavour.relpath(path, start=start)

    def relparts(self, path: str, base: str) -> Tuple[str, ...]:
        return self.parts(self.relpath(path, base))

    def as_posix(self, path: str) -> str:
        return path.replace(self.flavour.sep, posixpath.sep)



from typing import Dict, List, Optional, Type

from jmespath import exceptions, lexer
from jmespath.parser import ParsedResult, Parser
from jmespath.visitor import Options, TreeInterpreter

from ldb.jmespath.exceptions import (
    IdentifierTypeException,
    InvalidIdentifierExpression,
    MissingIdentifierException,
)
from ldb.typing import JSONDecoded, JSONObject


class FieldValidatedTreeInterpreter(  # pylint: disable=abstract-method
    TreeInterpreter,
):
    def visit_field(
        self,
        node: Dict[str, JSONDecoded],
        value: JSONDecoded,
    ) -> JSONDecoded:
        key: str = node["value"]  # type: ignore[assignment]
        try:
            return value[key]  # type: ignore[index,call-overload]
        except TypeError as exc:
            raise IdentifierTypeException(key, value) from exc
        except KeyError as exc:
            raise MissingIdentifierException(key) from exc


class CustomParser(Parser):
    def __init__(
        self,
        lookahead: int = 2,
        tree_interpreter_class: Type[TreeInterpreter] = TreeInterpreter,
    ) -> None:
        self._tree_interpreter_class = tree_interpreter_class
        super().__init__(lookahead=lookahead)

    def _parse(self, expression: str) -> "CustomParsedResult":
        self.tokenizer = (
            lexer.Lexer().tokenize(  # type: ignore[func-returns-value]
                expression,
            )
        )
        self._tokens = list(self.tokenizer)
        self._index = 0
        parsed = self._expression(  # type: ignore[attr-defined]
            binding_power=0,
        )
        if not self._current_token() == "eof":  # type: ignore[attr-defined]
            t = self._lookahead_token(0)  # type: ignore[attr-defined]
            raise exceptions.ParseError(
                t["start"],
                t["value"],
                t["type"],
                f"Unexpected token: {t['value']!s}",
            )
        return CustomParsedResult(
            expression,
            parsed,
            tree_interpreter_class=self._tree_interpreter_class,
        )


class CustomParsedResult(ParsedResult):
    def __init__(
        self,
        expression: str,
        parsed: JSONObject,
        tree_interpreter_class: Type[TreeInterpreter] = TreeInterpreter,
    ) -> None:
        self._tree_interpreter_class = tree_interpreter_class
        super().__init__(expression, parsed)

    def search(
        self,
        value: JSONDecoded,
        options: Optional[Options] = None,
    ) -> JSONDecoded:
        interpreter = self._tree_interpreter_class(options)
        result: JSONDecoded = interpreter.visit(  # type: ignore[no-untyped-call] # noqa: E501
            self.parsed,
            value,
        )
        return result


def parse_identifier_expression(key: str) -> List[str]:
    tokens = list(
        lexer.Lexer().tokenize(key),  # type: ignore[func-returns-value]
    )
    even = True
    result = []
    for t in tokens[:-1]:
        if even:
            if not t["type"] in ("unquoted_identifier", "quoted_identifier"):
                raise InvalidIdentifierExpression(key)
            result.append(t["value"])
        elif not t["type"] == "dot":
            raise InvalidIdentifierExpression(key)
        even = not even
    return result



from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import TYPE_CHECKING, Iterable

from ldb.add import sync
from ldb.cli_utils import add_data_obj_params

if TYPE_CHECKING:
    from argparse import _SubParsersAction


def sync_command(options: Namespace) -> None:
    sync(
        Path("."),
        options.paths,
        options.query_args,
    )


def add_parser(
    subparsers: "_SubParsersAction[ArgumentParser]",
    parents: Iterable[ArgumentParser],
) -> None:
    parser = subparsers.add_parser(
        "sync",
        parents=parents,
        help="Sync workspace to data objects under a certain path",
    )
    add_data_obj_params(parser, dest="query_args")
    parser.set_defaults(func=sync_command)



from contextlib import contextmanager
from typing import Any, Generator

from rich.progress import BarColumn, Progress, TextColumn, TimeRemainingColumn


@contextmanager
def get_progressbar(
    *args: Any,
    **kwargs: Any,
) -> Generator[Progress, None, None]:
    with Progress(
        TextColumn(
            "[bold blue]{task.completed}/{task.total}",
            justify="right",
        ),
        BarColumn(bar_width=None),
        "[progress.percentage]{task.percentage:>3.1f}%",
        "•",
        TimeRemainingColumn(),
        *args,
        **kwargs,
    ) as progress:
        yield progress



import operator
import re
from typing import Iterable, List, Optional, Union

from jmespath.exceptions import JMESPathTypeError

from ldb.jmespath.parser import parse_identifier_expression
from ldb.typing import JMESPathValue, JSONBinFunc, JSONDecoded

NumVec = Union[int, float, List[Union[int, float]]]
Arg = Union[JMESPathValue, NumVec]


def regex(string: str, pattern: str) -> bool:
    return re.search(pattern, string) is not None


def regex_match(string: str, pattern: str) -> Optional[str]:
    match = re.search(pattern, string)
    if match is None:
        return None
    return match.group()


def apply_bin_op(
    op: JSONBinFunc,
    x1: JMESPathValue,
    x2: JMESPathValue,
) -> Union[JMESPathValue, NumVec]:
    if isinstance(x1, Iterable):
        if isinstance(x2, Iterable):
            return [op(x1_j, x2_j) for x1_j, x2_j in zip(x1, x2)]
        return [op(x1_j, x2) for x1_j in x1]
    if isinstance(x2, Iterable):
        return [op(x1, x2_j) for x2_j in x2]
    return op(x1, x2)


def add(x1: NumVec, x2: NumVec) -> NumVec:
    return apply_bin_op(operator.add, x1, x2)  # type: ignore[return-value]


def sub(x1: NumVec, x2: NumVec) -> NumVec:
    return apply_bin_op(operator.sub, x1, x2)  # type: ignore[return-value]


def mul(x1: NumVec, x2: NumVec) -> NumVec:
    return apply_bin_op(operator.mul, x1, x2)  # type: ignore[return-value]


def div(x1: NumVec, x2: NumVec) -> NumVec:
    return apply_bin_op(operator.truediv, x1, x2)  # type: ignore[return-value]


def neg(x: NumVec) -> NumVec:
    if isinstance(x, Iterable):
        return [-x_i for x_i in x]
    return -x


def contains_all(
    subject: Union[List[JSONDecoded], str],
    searches: List[JSONDecoded],
) -> bool:
    return all(s in subject for s in searches)


def contains_any(
    subject: Union[List[JSONDecoded], str],
    searches: List[JSONDecoded],
) -> bool:
    return any(s in subject for s in searches)


def has_keys(value: JSONDecoded, *keys: str) -> bool:
    for key_exp in keys:
        node = value
        for key in parse_identifier_expression(key_exp):
            try:
                node = node[key]  # type: ignore[index,call-overload]
            except (KeyError, TypeError):
                return False
    return True


def get(
    value: JSONDecoded,
    key_exp: str,
    default: JSONDecoded = None,
) -> JSONDecoded:
    node = value
    # Optional args on variadic functions take the same types, so we
    # give `key_exp` a more general type signature and ensure it's a
    # string here
    if not isinstance(key_exp, str):
        raise JMESPathTypeError(
            "get",
            key_exp,
            type(key_exp).__name__,
            ["string"],
        )
    for key in parse_identifier_expression(key_exp):
        try:
            node = node[key]  # type: ignore[index,call-overload]
        except (KeyError, TypeError):
            return default
    return node


CUSTOM_FUNCTIONS = {
    "regex": (regex, ["string", "string"]),
    "regex_match": (regex_match, ["string", "string"]),
    "add": (add, ["number|array", "number|array"]),
    "sub": (sub, ["number|array", "number|array"]),
    "mul": (mul, ["number|array", "number|array"]),
    "div": (div, ["number|array", "number|array"]),
    "neg": (neg, ["number|array"]),
    "contains_all": (contains_all, ["array|string", "array"]),
    "contains_any": (contains_any, ["array|string", "array"]),
    "has_keys": (
        has_keys,
        ["boolean|array|object|null|string|number", "string", "*"],
    ),
    "get": (
        get,
        [
            "boolean|array|object|null|string|number",
            "boolean|array|object|null|string|number|expref",
            "*",
        ],
    ),
}



import json
import os
import random
from abc import ABC
from collections import abc, defaultdict
from dataclasses import asdict, dataclass, fields
from datetime import datetime
from glob import iglob
from itertools import tee
from pathlib import Path
from typing import (
    Any,
    Callable,
    Collection,
    DefaultDict,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Tuple,
    Union,
)

from funcy.objects import cached_property

from ldb.collections import LDBMappingCache
from ldb.exceptions import DatasetNotFoundError, LDBException
from ldb.iter_utils import take
from ldb.op_type import OpType
from ldb.path import InstanceDir
from ldb.query.search import (
    BoolSearchFunc,
    get_bool_search_func,
    get_no_tag_func,
    get_path_func,
    get_tag_func,
)
from ldb.typing import JSONDecoded, JSONObject
from ldb.utils import (
    format_dataset_identifier,
    format_datetime,
    get_hash_path,
    load_data_file,
    parse_datetime,
)

OpDef = Tuple[str, Union[str, int, float, List[str]]]
CollectionFunc = Callable[
    [Iterable[Tuple[str, str]]],
    Iterator[Tuple[str, str]],
]


@dataclass
class CommitInfo:
    created_by: str
    commit_time: datetime
    commit_message: str

    @classmethod
    def parse(cls, attr_dict: Dict[str, str]) -> "CommitInfo":
        attr_dict = attr_dict.copy()
        return cls(
            commit_time=parse_datetime(attr_dict.pop("commit_time")),
            **attr_dict,
        )

    def format(self) -> Dict[str, str]:
        attr_dict = asdict(self)
        return dict(
            commit_time=format_datetime(attr_dict.pop("commit_time")),
            **attr_dict,
        )


@dataclass
class DatasetVersion:
    version: int
    parent: str
    collection: str
    tags: List[str]
    commit_info: CommitInfo

    @classmethod
    def parse(cls, attr_dict: Dict[str, Any]) -> "DatasetVersion":
        attr_dict = attr_dict.copy()
        return cls(
            commit_info=CommitInfo.parse(attr_dict.pop("commit_info")),
            **attr_dict,
        )

    def format(self) -> Dict[str, Any]:
        attr_dict = {f.name: getattr(self, f.name) for f in fields(self)}
        return dict(
            commit_info=attr_dict.pop("commit_info").format(),
            tags=attr_dict.pop("tags").copy(),
            **attr_dict,
        )


@dataclass
class Dataset:
    name: str
    created_by: str
    created: datetime
    versions: List[str]

    @classmethod
    def parse(cls, attr_dict: Dict[str, Any]) -> "Dataset":
        attr_dict = attr_dict.copy()
        created = parse_datetime(attr_dict.pop("created"))
        return cls(created=created, **attr_dict)

    def format(self) -> Dict[str, Any]:
        attr_dict = asdict(self)
        created = format_datetime(attr_dict.pop("created"))
        return dict(created=created, **attr_dict)


def iter_collection_dir(collection_dir: Union[str, Path]) -> Iterator[str]:
    return iglob(os.path.join(collection_dir, "*", "*"))


def get_collection(
    ldb_dir: Path,
    dataset_version_hash: str,
) -> Dict[str, Optional[str]]:
    dataset_version_obj = DatasetVersion.parse(
        load_data_file(
            get_hash_path(
                ldb_dir / InstanceDir.DATASET_VERSIONS,
                dataset_version_hash,
            ),
        ),
    )
    return load_data_file(  # type: ignore[no-any-return]
        get_hash_path(
            ldb_dir / InstanceDir.COLLECTIONS,
            dataset_version_obj.collection,
        ),
    )


def get_collection_from_dataset_identifier(
    ldb_dir: Path,
    dataset_name: str,
    dataset_version: Optional[int] = None,
) -> Dict[str, Optional[str]]:
    dataset = get_dataset(ldb_dir, dataset_name)
    dataset_version_hash = get_dataset_version_hash(dataset, dataset_version)
    return get_collection(ldb_dir, dataset_version_hash)


def get_collection_dir_keys(
    collection_dir: Union[str, Path],
) -> Iterator[str]:
    for path in iter_collection_dir(collection_dir):
        parent, name = os.path.split(path)
        yield os.path.basename(parent) + name


def get_collection_dir_items(
    collection_dir: Path,
    is_workspace: bool = True,
) -> Iterator[Tuple[str, Optional[str]]]:
    annotation_hash_func = (
        get_workspace_collection_annotation_hash
        if is_workspace
        else get_root_collection_annotation_hash
    )
    for path in sorted(collection_dir.glob("*/*")):
        yield path.parent.name + path.name, annotation_hash_func(path)


def get_collection_size(
    collection_dir: Union[str, Path],
) -> int:
    return sum(1 for _ in iter_collection_dir(collection_dir))


def get_root_collection_annotation_hash(
    data_object_path: Path,
) -> Optional[str]:
    try:
        return (data_object_path / "current").read_text()
    except FileNotFoundError:
        return None


def get_workspace_collection_annotation_hash(
    data_object_path: Path,
) -> Optional[str]:
    return data_object_path.read_text() or None


def combine_collections(
    ldb_dir: Path,
    collections: List[Dict[str, Optional[str]]],
) -> Dict[str, str]:
    all_versions: DefaultDict[str, List[str]] = defaultdict(list)
    for collection in collections:
        for data_object_hash, annotation_hash in collection.items():
            lst = all_versions[data_object_hash]
            if annotation_hash:
                lst.append(annotation_hash)
    combined_collection = {}
    for data_object_hash, annotation_hashes in sorted(all_versions.items()):
        if len(annotation_hashes) > 1:
            annotation_dir = (
                get_hash_path(
                    ldb_dir / InstanceDir.DATA_OBJECT_INFO,
                    data_object_hash,
                )
                / "annotations"
            )
            latest_annotation_hash = max(
                (load_data_file(annotation_dir / h)["version"], h)
                for h in annotation_hashes
            )[1]
        elif annotation_hashes:
            latest_annotation_hash = annotation_hashes[0]
        else:
            latest_annotation_hash = ""
        combined_collection[data_object_hash] = latest_annotation_hash
    return combined_collection


def get_dataset(ldb_dir: Path, dataset_name: str) -> Dataset:
    try:
        return Dataset.parse(
            load_data_file(ldb_dir / InstanceDir.DATASETS / dataset_name),
        )
    except FileNotFoundError as exc:
        raise DatasetNotFoundError(
            f"Dataset not found with name {dataset_name!r}",
        ) from exc


def iter_dataset_dir(
    ldb_dir: Union[str, Path],
) -> Iterator[os.DirEntry]:  # type: ignore[type-arg]
    yield from os.scandir(os.path.join(ldb_dir, InstanceDir.DATASETS))


def iter_datasets(ldb_dir: Union[str, Path]) -> Iterator[Dataset]:
    for entry in iter_dataset_dir(ldb_dir):
        yield Dataset.parse(load_data_file(Path(entry.path)))


def get_dataset_version_hash(
    dataset: Dataset,
    dataset_version: Optional[int] = None,
) -> str:
    if not dataset_version:
        dataset_version = len(dataset.versions)
    try:
        return dataset.versions[dataset_version - 1]
    except IndexError as exc:
        dataset_identifier = format_dataset_identifier(
            dataset.name,
            dataset_version,
        )
        latest_dataset_identifier = format_dataset_identifier(
            dataset.name,
            len(dataset.versions),
        )
        raise LDBException(
            f"{dataset_identifier} does not exist\n"
            f"The latest version is {latest_dataset_identifier}",
        ) from exc


def get_annotation(ldb_dir: Path, annotation_hash: str) -> JSONDecoded:
    if not annotation_hash:
        return None
    user_annotation_file_path = (
        get_hash_path(
            ldb_dir / InstanceDir.ANNOTATIONS,
            annotation_hash,
        )
        / "user"
    )
    with open(user_annotation_file_path, encoding="utf-8") as f:
        data = f.read()
    return json.loads(data)  # type: ignore[no-any-return]


def get_annotations(
    ldb_dir: Path,
    annotation_hashes: Iterable[str],
) -> List[JSONDecoded]:
    annotations = []
    for annotation_hash in annotation_hashes:
        if annotation_hash:
            annotations.append(get_annotation(ldb_dir, annotation_hash))
        else:
            annotations.append(None)
    return annotations


def get_data_object_meta(
    ldb_dir: Path,
    data_object_hash: str,
) -> JSONObject:
    meta_file_path = (
        get_hash_path(
            ldb_dir / InstanceDir.DATA_OBJECT_INFO,
            data_object_hash,
        )
        / "meta"
    )
    meta: JSONObject = json.loads(meta_file_path.read_text())
    return meta


def get_data_object_metas(
    ldb_dir: Path,
    data_object_hashes: Iterable[str],
) -> List[JSONObject]:
    meta_objects = []
    for data_object_hash in data_object_hashes:
        meta_objects.append(get_data_object_meta(ldb_dir, data_object_hash))
    return meta_objects


class AnnotationCache(LDBMappingCache[str, JSONDecoded]):
    def get_new(self, key: str) -> JSONDecoded:
        return get_annotation(self.ldb_dir, key)


class DataObjectMetaCache(LDBMappingCache[str, JSONDecoded]):
    def get_new(self, key: str) -> JSONDecoded:
        return get_data_object_meta(self.ldb_dir, key)


class CollectionOperation(ABC):
    def apply(
        self,
        collection: Iterable[Tuple[str, str]],
    ) -> Iterator[Tuple[str, str]]:
        raise NotImplementedError


class Limit(CollectionOperation):
    def __init__(self, n: int) -> None:
        self.n = n

    def apply(
        self,
        collection: Iterable[Tuple[str, str]],
    ) -> Iterator[Tuple[str, str]]:
        return take(collection, n=self.n)


class Sample(CollectionOperation):
    def __init__(self, p: float) -> None:
        if not 0.0 <= p <= 1.0:
            raise ValueError(
                "sample probability must be in the range (0.0, 1.0) inclusive",
            )
        self.p = p

    def apply(
        self,
        collection: Iterable[Tuple[str, str]],
    ) -> Iterator[Tuple[str, str]]:
        for c in collection:
            if random.random() <= self.p:
                yield c


class Sort(CollectionOperation):
    def __init__(self, ldb_dir: Path, proc_args: List[str]) -> None:
        self.ldb_dir = ldb_dir
        self.proc_args = proc_args

    def apply(
        self,
        collection: Iterable[Tuple[str, str]],
    ) -> Iterator[Tuple[str, str]]:
        from ldb.pipe import (  # pylint: disable=import-outside-toplevel
            sort_collection,
        )

        return sort_collection(self.ldb_dir, collection, self.proc_args)


class Query(CollectionOperation):
    def __init__(
        self,
        ldb_dir: Path,
        cache: LDBMappingCache[str, Any],
        search: BoolSearchFunc,
    ) -> None:
        self.ldb_dir = ldb_dir
        self.cache = cache
        self.search = search

    def apply(
        self,
        collection: Iterable[Tuple[str, str]],
    ) -> Iterator[Tuple[str, str]]:
        collection1, collection2 = tee(collection)
        for (data_object_hash, annotation_hash), keep in zip(
            collection1,
            self.search(self.get_search_input(collection2)),
        ):
            if keep:
                yield data_object_hash, annotation_hash

    def get_search_input(
        self,
        collection: Iterable[Tuple[str, str]],
    ) -> Iterable[JSONDecoded]:
        raise NotImplementedError


class AnnotationQuery(Query):
    def get_search_input(
        self,
        collection: Iterable[Tuple[str, str]],
    ) -> Iterator[JSONDecoded]:
        for _, annot_hash in collection:
            yield self.cache[annot_hash]


class FileQuery(Query):
    def get_search_input(
        self,
        collection: Iterable[Tuple[str, str]],
    ) -> Iterator[JSONDecoded]:
        for data_obj_hash, _ in collection:
            yield self.cache[data_obj_hash]


class TagQuery(FileQuery):
    def get_search_input(
        self,
        collection: Iterable[Tuple[str, str]],
    ) -> Iterator[List[str]]:
        for data_obj_meta in super().get_search_input(collection):
            yield data_obj_meta["tags"]  # type: ignore[index,call-overload,misc] # noqa: E501


class PathQuery(FileQuery):
    def get_search_input(
        self,
        collection: Iterable[Tuple[str, str]],
    ) -> Iterator[List[str]]:
        for data_obj_meta in super().get_search_input(collection):
            yield [info["path"] for info in data_obj_meta["alternate_paths"]]  # type: ignore[index,call-overload] # noqa: E501


class PipelineData:
    def __init__(self, ldb_dir: Path) -> None:
        self.ldb_dir = ldb_dir

    @cached_property
    def data_object_metas(self) -> DataObjectMetaCache:
        return DataObjectMetaCache(self.ldb_dir)

    @cached_property
    def annotations(self) -> AnnotationCache:
        return AnnotationCache(self.ldb_dir)


class PipelineBuilder:
    def __init__(
        self,
        ldb_dir: Path,
        data: Optional[PipelineData] = None,
    ) -> None:
        self.ldb_dir = ldb_dir
        if data is None:
            self.data = PipelineData(ldb_dir)
        else:
            self.data = data

    def build(
        self,
        op_defs: Iterable[OpDef],
        warn: bool = True,
    ) -> List[CollectionFunc]:
        ops = []
        for op_type, arg in op_defs:
            op: CollectionFunc
            if op_type == OpType.ANNOTATION_QUERY:
                assert isinstance(arg, str)
                op = self.annotation_query(arg, True, warn)
            elif op_type == OpType.JP_ANNOTATION_QUERY:
                assert isinstance(arg, str)
                op = self.annotation_query(arg, False, warn)
            elif op_type == OpType.FILE_QUERY:
                assert isinstance(arg, str)
                op = self.file_query(arg, warn)
            elif op_type == OpType.TAG_QUERY:
                assert isinstance(arg, abc.Collection)
                op = self.tag_query(arg)
            elif op_type == OpType.NO_TAG_QUERY:
                assert isinstance(arg, abc.Collection)
                op = self.no_tag_query(arg)
            elif op_type == OpType.PATH_QUERY:
                assert isinstance(arg, str)
                op = self.path_query(arg)
            elif op_type == OpType.LIMIT:
                assert isinstance(arg, int)
                op = Limit(arg).apply
            elif op_type == OpType.SAMPLE:
                assert isinstance(arg, float)
                op = Sample(arg).apply
            elif op_type == OpType.PIPE:
                assert isinstance(arg, list)
                op = Sort(self.ldb_dir, arg).apply
            else:
                raise ValueError(f"Unknown op type: {op_type}")
            ops.append(op)
        return ops

    def annotation_query(
        self,
        search: str,
        use_custom: bool,
        warn: bool,
    ) -> CollectionFunc:
        return AnnotationQuery(
            self.ldb_dir,
            self.data.annotations,
            get_bool_search_func(search, use_custom=use_custom, warn=warn),
        ).apply

    def file_query(self, search: str, warn: bool) -> CollectionFunc:
        return FileQuery(
            self.ldb_dir,
            self.data.data_object_metas,
            get_bool_search_func(search, warn=warn),
        ).apply

    def tag_query(self, tag: Collection[str]) -> CollectionFunc:
        return TagQuery(
            self.ldb_dir,
            self.data.data_object_metas,
            get_tag_func(tag),
        ).apply

    def no_tag_query(self, tag: Collection[str]) -> CollectionFunc:
        return TagQuery(
            self.ldb_dir,
            self.data.data_object_metas,
            get_no_tag_func(tag),
        ).apply

    def path_query(self, pattern: str) -> CollectionFunc:
        return PathQuery(
            self.ldb_dir,
            self.data.data_object_metas,
            get_path_func(pattern),
        ).apply


class Pipeline:
    def __init__(self, ops: Iterable[CollectionFunc]) -> None:
        self.ops = ops

    @classmethod
    def from_defs(
        cls,
        ldb_dir: Path,
        op_defs: Iterable[OpDef],
        data: Optional[PipelineData] = None,
        warn: bool = True,
    ) -> "Pipeline":
        return cls(
            PipelineBuilder(ldb_dir, data=data).build(op_defs, warn=warn),
        )

    def run(
        self,
        collection: Iterable[Tuple[str, str]],
    ) -> Iterator[Tuple[str, str]]:
        for op in self.ops:
            collection = op(collection)
        return iter(collection)


def apply_queries(
    ldb_dir: Path,
    data_object_hashes: Iterable[str],
    annotation_hashes: Iterable[str],
    op_defs: Iterable[OpDef],
    warn: bool = True,
) -> Iterator[Tuple[str, str]]:
    collection = zip(data_object_hashes, annotation_hashes)
    return apply_queries_to_collection(ldb_dir, collection, op_defs, warn=warn)


def apply_queries_to_collection(
    ldb_dir: Path,
    collection: Iterable[Tuple[str, str]],
    op_defs: Iterable[OpDef],
    warn: bool = True,
) -> Iterator[Tuple[str, str]]:
    """
    Filter the given collection by the operations in `collection_ops`.
    """
    return Pipeline.from_defs(ldb_dir, op_defs, warn=warn).run(collection)



import os
import shutil
from pathlib import Path
from typing import Dict, List, Optional, Tuple

from ldb.dataset import Dataset, DatasetVersion
from ldb.exceptions import LDBException
from ldb.path import InstanceDir, WorkspacePath
from ldb.utils import (
    DATASET_PREFIX,
    ROOT,
    current_time,
    format_dataset_identifier,
    get_hash_path,
    json_dumps,
    load_data_file,
    parse_dataset_identifier,
    write_data_file,
)
from ldb.workspace import (
    WorkspaceDataset,
    ensure_path_is_empty_workspace,
    workspace_dataset_is_clean,
)


def stage(
    ldb_dir: Path,
    dataset_identifier: str,
    workspace_path: Path,
    force: bool = False,
) -> None:
    ds_name, ds_version_num = parse_dataset_identifier(dataset_identifier)
    if ds_name == ROOT:
        raise ValueError(
            f"Cannot stage dataset named {DATASET_PREFIX}{ds_name}",
        )
    workspace_path = Path(os.path.normpath(workspace_path))
    if workspace_path.is_dir():
        if not force:
            try:
                curr_workspace_ds_obj = WorkspaceDataset.parse(
                    load_data_file(
                        workspace_path / WorkspacePath.DATASET,
                    ),
                )
            except FileNotFoundError:
                pass
            else:
                if not workspace_dataset_is_clean(
                    ldb_dir,
                    curr_workspace_ds_obj,
                    workspace_path,
                ):
                    raise LDBException(
                        "Unsaved changes to workspace dataset.\n"
                        "Commit changes or use the --force option to "
                        "overwrite them.",
                    )
        ensure_path_is_empty_workspace(workspace_path, force)
    workspace_ds_obj = WorkspaceDataset(
        dataset_name=ds_name,
        staged_time=current_time(),
        parent="",
        tags=[],
    )
    try:
        dataset_obj = Dataset.parse(
            load_data_file(ldb_dir / InstanceDir.DATASETS / ds_name),
        )
    except FileNotFoundError as exc:
        if ds_version_num is not None:
            ds_ident = format_dataset_identifier(ds_name)
            raise LDBException(
                f"{dataset_identifier} does not exist\n"
                f"To stage a new dataset, use {ds_ident}",
            ) from exc
        collection_obj = None
        message = (
            f"Staged new dataset {dataset_identifier} "
            f"at {os.fspath(workspace_path)!r}"
        )
    else:
        if ds_version_num is None:
            dataset_version_hash = dataset_obj.versions[-1]
            ds_version_num = len(dataset_obj.versions)
        else:
            try:
                dataset_version_hash = dataset_obj.versions[ds_version_num - 1]
            except IndexError as exc:
                latest_dataset = format_dataset_identifier(
                    ds_name,
                    len(dataset_obj.versions),
                )
                raise LDBException(
                    f"{dataset_identifier} does not exist\n"
                    f"The latest version is {latest_dataset}",
                ) from exc
        dataset_version_obj = DatasetVersion.parse(
            load_data_file(
                get_hash_path(
                    ldb_dir / InstanceDir.DATASET_VERSIONS,
                    dataset_version_hash,
                ),
            ),
        )
        workspace_ds_obj.parent = dataset_version_hash
        workspace_ds_obj.tags = dataset_version_obj.tags.copy()
        collection_obj = load_data_file(
            get_hash_path(
                ldb_dir / InstanceDir.COLLECTIONS,
                dataset_version_obj.collection,
            ),
        )
        curr_dataset_ident = format_dataset_identifier(
            ds_name,
            ds_version_num,
        )
        message = (
            f"Staged {curr_dataset_ident} at {os.fspath(workspace_path)!r}"
        )
    stage_workspace(workspace_path, workspace_ds_obj, collection_obj)
    print(message)


def stage_workspace(
    workspace_path: Path,
    workspace_ds_obj: WorkspaceDataset,
    collection_obj: Optional[Dict[str, Optional[str]]] = None,
) -> None:
    collection_path = workspace_path / WorkspacePath.COLLECTION
    workspace_ds_bytes = json_dumps(workspace_ds_obj.format()).encode()
    if collection_obj:
        collection_path_data = get_workspace_collection_path_data(
            collection_path,
            collection_obj,
        )
        collection_path.mkdir(parents=True, exist_ok=True)
        write_workspace_collection(collection_path, collection_path_data)
    else:
        if collection_path.exists():
            shutil.rmtree(collection_path)
        collection_path.mkdir(parents=True)
    write_data_file(
        workspace_path / WorkspacePath.DATASET,
        workspace_ds_bytes,
    )


def get_workspace_collection_path_data(
    path: Path,
    collection_obj: Dict[str, Optional[str]],
) -> List[Tuple[Path, str]]:
    return [
        (get_hash_path(path, data_object_hash), annotation_hash or "")
        for data_object_hash, annotation_hash in collection_obj.items()
    ]


def write_workspace_collection(
    collection_path: Path,
    path_data: List[Tuple[Path, str]],
) -> None:
    for path, data in path_data:
        path.parent.mkdir(exist_ok=True)
        path.write_text(data)
    path_set = {d[0] for d in path_data}
    for path in collection_path.glob("*/*"):
        if path not in path_set:
            path.unlink()
    path_parent_set = {p.parent for p in path_set}
    for path in collection_path.glob("*"):
        if path not in path_parent_set:
            path.rmdir()



from argparse import ArgumentParser, Namespace
from typing import TYPE_CHECKING, Iterable

import shtab

from ldb.config import get_ldb_dir
from ldb.core import init
from ldb.path import Filename
from ldb.storage import add_storage, create_storage_location

if TYPE_CHECKING:
    from argparse import _SubParsersAction


def add_storage_command(options: Namespace) -> None:
    ldb_dir = get_ldb_dir()
    if not ldb_dir.is_dir():
        ldb_dir = init(ldb_dir)
    storage_location = create_storage_location(
        path=options.path,
        read_and_add=options.read_add,
        options=dict(options.fs_options),
    )
    add_storage(
        ldb_dir / Filename.STORAGE,
        storage_location,
        force=options.force,
    )


def add_parser(
    subparsers: "_SubParsersAction[ArgumentParser]",
    parents: Iterable[ArgumentParser],
) -> None:
    parser = subparsers.add_parser(
        "add-storage",
        parents=parents,
        help="Add a storage location",
    )
    parser.add_argument(
        "-a",
        "--read-add",
        action="store_true",
        default=False,
        help="Use this location for adding objects",
    )
    parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        default=False,
        help="Replace a subdirectory with one of its parents",
    )
    parser.add_argument(
        "-o",
        "--option",
        metavar=("<key>", "<value>"),
        default=[],
        action="append",
        dest="fs_options",
        nargs=2,
        help="Additional filesystem option. May be used multiple times",
    )
    parser.add_argument(  # type: ignore[attr-defined]
        "path",
        metavar="<path>",
        help="The path to the location",
    ).complete = shtab.DIRECTORY
    parser.set_defaults(func=add_storage_command)



from typing import Callable, Optional, Sequence, Union

import clip
import torch
from PIL import Image
from PIL.PngImagePlugin import PngImageFile
from torch.utils.data import DataLoader, Dataset
from torchvision.transforms import Compose, ToTensor
from tqdm import tqdm


class ImageOnlyDataset(Dataset[torch.Tensor]):
    def __init__(
        self,
        file_paths: Sequence[str],
        transform: Optional[Callable[[PngImageFile], torch.Tensor]] = None,
    ) -> None:
        self.file_paths = file_paths
        self.transform = transform if transform is not None else ToTensor()

    def __len__(self) -> int:
        return len(self.file_paths)

    def __getitem__(self, index: int) -> torch.Tensor:
        image: PngImageFile
        image = Image.open(self.file_paths[index])  # type: ignore[assignment]
        image_tensor: torch.Tensor = self.transform(image)
        return image_tensor


def get_device() -> str:
    return "cuda" if torch.cuda.is_available() else "cpu"


def get_image_dataset_features(
    model: clip.model.CLIP,
    device: str,
    dataset: Dataset[torch.Tensor],
) -> torch.Tensor:
    all_features = []
    with torch.no_grad():  # type: ignore[no-untyped-call]
        for images in tqdm(DataLoader(dataset, batch_size=64)):
            features = model.encode_image(images.to(device))
            all_features.append(features)
    return torch.cat(
        all_features,
    ).cpu()  # pylint: disable=no-member # noqa: E501


def get_image_features(
    model: torch.Tensor,
    preprocess: Compose,
    device: str,
    file_paths: Sequence[str],
) -> torch.Tensor:
    dataset = ImageOnlyDataset(file_paths, transform=preprocess)
    return get_image_dataset_features(model, device, dataset)


def get_text_features(
    model: clip.model.CLIP,
    device: str,
    text: Union[str, Sequence[str]],
) -> torch.Tensor:
    tokens = clip.tokenize(text).to(device)
    with torch.no_grad():  # type: ignore[no-untyped-call]
        text_features: torch.Tensor = model.encode_text(tokens)
    return text_features


def text_similarity(
    text: str,
    file_paths: Sequence[str],
    model_name: Optional[str] = None,
) -> torch.Tensor:
    """
    Return similarities between a text and a sequence of images.

    The similarities are given as a Tensor containing the sequence of
    cosine similarities between an encoding of `text` and each encoding
    of the images at `file_paths`.
    """
    if model_name is None:
        model_name = "ViT-B/32"
    device = get_device()

    model, preprocess = clip.load(
        model_name,
        device=device,
    )

    text_features = get_text_features(model, device, text)
    image_features = get_image_features(model, preprocess, device, file_paths)

    image_features /= image_features.norm(  # type: ignore[no-untyped-call]
        dim=-1,
        keepdim=True,
    )
    text_features /= text_features.norm(  # type: ignore[no-untyped-call]
        dim=-1,
        keepdim=True,
    )
    return (text_features @ image_features.T)[0]


def image_similarity(
    image_filepath: str,
    file_paths: Sequence[str],
    model_name: Optional[str] = None,
) -> torch.Tensor:
    """
    Return similarities between a target image and a sequence of images.

    The similarities are given as a Tensor containing the sequence of
    cosine similarities between an encoding of the image at
    `image_filepath` each encoding of the images at `file_paths`.
    """
    if model_name is None:
        model_name = "ViT-B/32"
    device = get_device()
    model, preprocess = clip.load(model_name, device=device)

    image_features = get_image_features(
        model,
        preprocess,
        device,
        [image_filepath, *file_paths],
    )

    image_features /= image_features.norm(  # type: ignore[no-untyped-call]
        dim=-1,
        keepdim=True,
    )
    return (image_features[:1] @ image_features[1:].T)[0]



import hashlib
import json
import os
import random
import re
import stat
import string
from contextlib import contextmanager
from datetime import datetime, timezone
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Iterator,
    Optional,
    Tuple,
    TypeVar,
    Union,
    overload,
)

from fsspec.spec import AbstractFileSystem

from ldb.fs.utils import first_protocol

if TYPE_CHECKING:
    from _typeshed import SupportsGetItem, SupportsRead

DATASET_PREFIX = "ds:"
WORKSPACE_DATASET_PREFIX = "ws:"
DATA_OBJ_ID_PREFIX = "id:"
ROOT = "root"
CHUNK_SIZE = 2**20
HASH_DIR_SPLIT_POINT = 3
UNIQUE_ID_ALPHABET = string.ascii_lowercase + string.digits
DATA_OBJ_ID_PATTERN = "^(?:id:)?([0-9a-f]{32})$"

_KT_contra = TypeVar("_KT_contra", contravariant=True)
_VT_co = TypeVar("_VT_co", covariant=True)
_T = TypeVar("_T")


@overload
def get_first(
    container: "SupportsGetItem[_KT_contra, _VT_co]",
    *keys: _KT_contra,
) -> Optional[_VT_co]:
    ...


@overload
def get_first(
    container: "SupportsGetItem[_KT_contra, _VT_co]",
    *key: _KT_contra,
    default: _T,
) -> Union[_VT_co, _T]:
    ...


def get_first(
    container: "SupportsGetItem[_KT_contra, _VT_co]",
    *keys: _KT_contra,
    default: Optional[_T] = None,
) -> Union[_VT_co, Optional[_T]]:
    for key in keys:
        try:
            return container[key]
        except LookupError:
            pass
    return default


def json_dumps(obj: Any, **kwargs: Any) -> str:
    return json.dumps(obj, sort_keys=True, **kwargs)


def write_data_file(
    file_path: Path,
    data: bytes,
    overwrite_existing: bool = True,
) -> None:
    if overwrite_existing or not file_path.exists():
        file_path.parent.mkdir(parents=True, exist_ok=True)
        with file_path.open("wb") as file:
            file.write(data)


def get_etag_md5_match(etag: str) -> str:
    # The e-tag will be a json string, so look for double quotes
    md5_hash_match = re.match(r"(?i)\"([a-f\d]{32})\"", etag)
    if md5_hash_match is None:
        return ""
    return md5_hash_match.group(1)


def get_file_hash(fs: AbstractFileSystem, path: str) -> str:
    if first_protocol(fs.protocol) in ("s3", "s3a"):
        return get_etag_md5_match(fs.info(path).get("ETag", "")) or hash_file(
            fs,
            path,
        )
    return hash_file(fs, path)


def hash_file(fs: AbstractFileSystem, path: str) -> str:
    hash_obj = hashlib.md5()
    with fs.open(path, "rb") as file:
        for chunk in iter_chunks(file):
            hash_obj.update(chunk)
    return hash_obj.hexdigest()


def hash_data(data: bytes) -> str:
    return hashlib.md5(data).hexdigest()


def iter_chunks(
    file: "SupportsRead[bytes]",
    chunk_size: int = CHUNK_SIZE,
) -> Iterator[bytes]:
    data = file.read(chunk_size)
    while data:
        yield data
        data = file.read(chunk_size)


def get_hash_path(base_dir: Path, hash_str: str) -> Path:
    return (
        base_dir
        / hash_str[:HASH_DIR_SPLIT_POINT]
        / hash_str[HASH_DIR_SPLIT_POINT:]
    )


def format_datetime(dt_obj: datetime) -> str:
    return dt_obj.astimezone(timezone.utc).isoformat(" ")


def parse_datetime(dt_str: str) -> datetime:
    return datetime.fromisoformat(dt_str).astimezone(timezone.utc)


def timestamp_to_datetime(timestamp: float) -> datetime:
    return datetime.fromtimestamp(timestamp, timezone.utc)


def current_time() -> datetime:
    return datetime.now(timezone.utc)


def load_data_file(path: Path) -> Any:
    with path.open() as file:
        return json.load(file)


def get_filetype(path: str) -> str:
    return os.path.splitext(path)[1].lstrip(".")


def parse_dataset_identifier(
    dataset_identifier: str,
) -> Tuple[str, Optional[int]]:
    match = re.search(
        rf"^{DATASET_PREFIX}([A-Za-z0-9_-]+)(?:\.v(\d+))?$",
        dataset_identifier,
    )
    if match is None:
        raise ValueError(
            'dataset identifier must be in the form "ds:name" or '
            '"ds:name.vN"\n'
            'where "name" can contain characters in the group '
            "[A-Za-z0-9_-]\n"
            'and ".vN" denotes the version number (e.g. ".v1")',
        )
    name, version = match.groups()
    return name, int(version) if version is not None else None


def format_dataset_identifier(
    name: str,
    version: Optional[int] = None,
) -> str:
    version_suffix = f".v{version}" if version else ""
    return f"{DATASET_PREFIX}{name}{version_suffix}"


def get_fsspec_path_suffix(path: str) -> str:
    match = re.search(r"\.[^/]*/*$", path)
    if match is None:
        return ""
    return match.group()


def parse_data_object_hash_identifier(hash_identifier: str) -> str:
    match = re.search(DATA_OBJ_ID_PATTERN, hash_identifier)
    if match is None:
        raise ValueError(
            "hash_identifier must be 32 hexadecimal characters, optionally "
            f"prefixed with {DATA_OBJ_ID_PREFIX}",
        )
    return match.group(1)


def unique_id(n: int = 8) -> str:
    return "".join(random.choices(UNIQUE_ID_ALPHABET, k=n))


@contextmanager
def chdir(path: Union[str, bytes, Path]) -> Iterator[None]:
    old = os.getcwd()
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(old)


def chmod_minus_x(path: Union[str, bytes, Path]) -> int:
    """
    Turn off the executable bit for everyone.

    Equivalent the unix `chmod -x`
    """
    mode = os.stat(path).st_mode & ~(
        stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH
    )
    os.chmod(path, mode)
    return mode



from argparse import ArgumentParser, Namespace
from typing import TYPE_CHECKING, Iterable

from ldb.cli_utils import add_target_dir_argument
from ldb.config import get_ldb_dir
from ldb.core import init_quickstart
from ldb.stage import stage

if TYPE_CHECKING:
    from argparse import _SubParsersAction


def stage_command(options: Namespace) -> None:
    ldb_dir = get_ldb_dir()
    if not ldb_dir.is_dir():
        ldb_dir = init_quickstart()
    stage(
        ldb_dir,
        options.dataset,
        options.target_dir,
        options.force,
    )


def add_parser(
    subparsers: "_SubParsersAction[ArgumentParser]",
    parents: Iterable[ArgumentParser],
) -> None:
    parser = subparsers.add_parser(
        "stage",
        parents=parents,
        help="Stage a dataset in the current workspace",
    )
    parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        default=False,
        help="Overwrite an unsaved dataset",
    )
    add_target_dir_argument(parser)
    parser.add_argument(
        "dataset",
        metavar="<dataset>",
        help="Name of the dataset to stage",
    )
    parser.set_defaults(func=stage_command)



from pathlib import PurePath


class DirName:
    GLOBAL_BASE = ".ldb"


class GlobalDir:
    DEFAULT_INSTANCE = PurePath("private_instance")
    DEFAULT_READ_ADD_STORAGE = PurePath("read_add_storage")


class InstanceDir:
    DATA_OBJECT_INFO = PurePath("data_object_info")
    DATASETS = PurePath("datasets")
    OBJECTS = PurePath("objects")
    ANNOTATIONS = OBJECTS / "annotations"
    COLLECTIONS = OBJECTS / "collections"
    DATASET_VERSIONS = OBJECTS / "dataset_versions"
    USER_FUNCTIONS = PurePath("custom_code") / "ldb_user_functions"
    USER_FILTERS = PurePath("custom_code") / "ldb_user_filters"


class WorkspacePath:
    BASE = PurePath(".ldb_workspace")
    COLLECTION = BASE / "collection"
    DATASET = BASE / "workspace_dataset"
    TMP = BASE / "tmp"


INSTANCE_DIRS = (
    InstanceDir.DATA_OBJECT_INFO,
    InstanceDir.DATASETS,
    InstanceDir.OBJECTS,
    InstanceDir.ANNOTATIONS,
    InstanceDir.COLLECTIONS,
    InstanceDir.DATASET_VERSIONS,
    InstanceDir.USER_FUNCTIONS,
    InstanceDir.USER_FILTERS,
)


class Filename:
    CONFIG = "config"
    STORAGE = "storage"



from abc import ABC
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path
from typing import (
    Any,
    Collection,
    Dict,
    Iterable,
    List,
    Mapping,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
)

from fsspec.spec import AbstractFileSystem
from funcy.objects import cached_property

from ldb.dataset import get_collection_dir_keys
from ldb.exceptions import IndexingException, LDBException
from ldb.index.utils import (
    DEFAULT_CONFIG,
    INDEXED_EPHEMERAL_CONFIG,
    AnnotationMeta,
    DataObjectMeta,
    DataToWrite,
    FileSystemPath,
    FSPathsMapping,
    IndexingJobMapping,
    construct_annotation_meta,
    construct_data_object_meta,
    copy_to_read_add_storage,
    data_object_path_to_annotation_path,
    expand_indexing_paths,
    get_annotation_content,
    group_indexing_paths_by_type,
    separate_indexed_files,
    separate_local_and_cloud_files,
    separate_storage_and_non_storage_files,
    validate_locations_in_storage,
)
from ldb.path import InstanceDir
from ldb.storage import StorageLocation
from ldb.typing import JSONDecoded, JSONObject
from ldb.utils import (
    current_time,
    format_datetime,
    get_file_hash,
    get_hash_path,
    hash_data,
    json_dumps,
    load_data_file,
    timestamp_to_datetime,
    write_data_file,
)

ENDING_DOUBLE_STAR_RE = r"(?:/+\*\*)+/*$"


class IndexedObjectResult(NamedTuple):
    found_data_object: bool
    found_annotation: bool
    new_data_object: bool
    new_annotation: bool
    data_object_hash: str


@dataclass
class IndexingResult:
    num_found_data_objects: int = 0
    num_found_annotations: int = 0
    num_new_data_objects: int = 0
    num_new_annotations: int = 0
    data_object_hashes: List[str] = field(default_factory=list)

    def append(self, item: IndexedObjectResult) -> None:
        self.num_found_data_objects += item.found_data_object
        self.num_found_annotations += item.found_annotation
        self.num_new_data_objects += item.new_data_object
        self.num_new_annotations += item.new_annotation
        self.data_object_hashes.append(item.data_object_hash)

    def summary(self, finished: bool = True) -> str:
        if finished:
            heading = "Finished indexing"
        else:
            heading = "Unable to finish indexing. Partial indexing results"
        return (
            f"\n{heading}:\n"
            f"  Found data objects: {self.num_found_data_objects:9d}\n"
            f"  Found annotations:  {self.num_found_annotations:9d}\n"
            f"  New data objects:   {self.num_new_data_objects:9d}\n"
            f"  New annotations:    {self.num_new_annotations:9d}"
        )


class Preprocessor:
    def __init__(
        self,
        paths: Sequence[str],
        storage_locations: Sequence[StorageLocation],
    ) -> None:
        self.paths: List[str] = list(paths)
        self.storage_locations = storage_locations

    def get_storage_files(self) -> FSPathsMapping:
        return expand_indexing_paths(
            self.paths,
            self.storage_locations,
            default_format=True,
        )

    @cached_property
    def files_by_type(self) -> Tuple[FSPathsMapping, FSPathsMapping]:
        data_obj_files = {}
        annot_files = {}
        for fs, fs_paths in self.get_storage_files().items():
            data_obj_files[fs], annot_files[fs] = group_indexing_paths_by_type(
                fs,
                fs_paths,
            )
        return data_obj_files, annot_files

    @cached_property
    def data_object_paths(self) -> FSPathsMapping:
        return self.files_by_type[0]

    @cached_property
    def annotation_paths(self) -> FSPathsMapping:
        return self.files_by_type[1]


class Indexer(ABC):
    def __init__(
        self,
        ldb_dir: Path,
        preprocessor: Preprocessor,
        tags: Collection[str] = (),
    ) -> None:
        self.ldb_dir = ldb_dir
        self.preprocessor = preprocessor
        self.tags = tags
        self.result = IndexingResult()
        self.hashes: Dict[AbstractFileSystem, Dict[str, str]] = {}

    def index(self) -> None:
        try:
            self._index()
        except Exception:
            print(self.result.summary(finished=False), "\n", sep="")
            raise

    def _index(self) -> None:
        raise NotImplementedError

    def process_files(self) -> Any:
        raise NotImplementedError


class PairIndexer(Indexer):
    def __init__(
        self,
        ldb_dir: Path,
        preprocessor: Preprocessor,
        read_any_cloud_location: bool,
        strict_format: bool,
        tags: Collection[str] = (),
    ) -> None:
        super().__init__(ldb_dir, preprocessor, tags)
        self.read_any_cloud_location = read_any_cloud_location
        self.strict_format = strict_format
        self.old_to_new_files: Dict[
            AbstractFileSystem,
            Dict[str, FileSystemPath],
        ] = {}
        self.old_to_new_annot_files: Dict[
            AbstractFileSystem,
            Dict[str, FileSystemPath],
        ] = {}

    def _index(self) -> None:
        indexing_jobs, annotation_paths = self.process_files()
        self.index_files(
            indexing_jobs,
            annotation_paths,
        )

    def process_files(
        self,
    ) -> Tuple[IndexingJobMapping, FSPathsMapping]:
        storage_locations = self.preprocessor.storage_locations
        local_files, cloud_files = separate_local_and_cloud_files(
            self.preprocessor.data_object_paths,
        )
        if not self.read_any_cloud_location:
            validate_locations_in_storage(cloud_files, storage_locations)
        (
            local_storage_files,
            ephemeral_files,
        ) = separate_storage_and_non_storage_files(
            local_files,
            storage_locations,
        )

        for fs, paths in ephemeral_files.items():
            fs_hashes = self.hashes.setdefault(fs, {})
            for path in paths:
                fs_hashes[path] = get_file_hash(fs, path)

        existing_hashes = set(
            get_collection_dir_keys(
                self.ldb_dir / InstanceDir.DATA_OBJECT_INFO,
            ),
        )

        indexed_ephemeral_files, ephemeral_files = separate_indexed_files(
            existing_hashes,
            self.hashes,
            ephemeral_files,
        )
        files = {
            fs: cloud_files.get(fs, []) + local_storage_files.get(fs, [])
            for fs in cloud_files.keys() | local_storage_files.keys()
        }
        annotation_paths = {
            fs: paths.copy()
            for fs, paths in self.preprocessor.annotation_paths.items()
        }

        if ephemeral_files:
            read_add_location = next(
                (loc for loc in storage_locations if loc.read_and_add),
                None,
            )
            if read_add_location is None:
                raise LDBException(
                    "No read-add storage configured. See 'ldb add-storage -h'",
                )
            (
                self.old_to_new_files,
                self.old_to_new_annot_files,
            ) = copy_to_read_add_storage(
                ephemeral_files,
                self.preprocessor.annotation_paths,
                read_add_location,
                self.hashes,
                strict_format=self.strict_format,
            )

            for fs, fs_paths in self.old_to_new_annot_files.items():
                for _, new in fs_paths.items():
                    annotation_paths.setdefault(new.fs, []).append(new.path)
            for fs, fs_paths in self.old_to_new_files.items():
                for _, new in fs_paths.items():
                    files.setdefault(new.fs, []).append(new.path)
        indexing_jobs: IndexingJobMapping = {}
        for config_type, fs_jobs in [
            (DEFAULT_CONFIG, files),
            (INDEXED_EPHEMERAL_CONFIG, indexed_ephemeral_files),
        ]:
            for fs, job in fs_jobs.items():
                indexing_jobs.setdefault(fs, []).append((config_type, job))
        return (
            indexing_jobs,
            annotation_paths,
        )

    def index_files(
        self,
        indexing_jobs: IndexingJobMapping,
        annotation_paths: FSPathsMapping,
    ) -> None:
        for fs, jobs in indexing_jobs.items():
            fs_annotation_paths = set(annotation_paths.get(fs, []))
            for config, path_seq in jobs:
                for data_object_path in path_seq:
                    annotation_path = data_object_path_to_annotation_path(
                        data_object_path,
                    )
                    try:
                        fs_annotation_paths.remove(
                            annotation_path,
                        )
                    except KeyError:
                        annotation_path = ""
                    if not self.strict_format or annotation_path:
                        obj_result = self.index_single_pair(
                            fs,
                            data_object_path,
                            config.save_data_object_path_info,
                            annotation_path,
                            self.tags,
                        )
                        self.result.append(obj_result)

    def index_single_pair(
        self,
        fs: AbstractFileSystem,
        data_object_path: str,
        save_data_object_path_info: bool,
        annotation_path: str,
        tags: Collection[str],
    ) -> IndexedObjectResult:
        return PairIndexingItem(
            self.ldb_dir,
            current_time(),
            tags,
            FileSystemPath(fs, data_object_path),
            save_data_object_path_info,
            self.hashes,
            FileSystemPath(fs, annotation_path) if annotation_path else None,
        ).index_data()


@dataclass
class IndexingItem(ABC):
    ldb_dir: Path
    curr_time: datetime
    _to_write: List[DataToWrite] = field(
        init=False,
        default_factory=list,
    )
    tags: Collection[str]

    @cached_property
    def current_timestamp(self) -> str:
        return format_datetime(self.curr_time)

    @cached_property
    def data_object_dir(self) -> Path:
        return get_hash_path(
            self.ldb_dir / InstanceDir.DATA_OBJECT_INFO,
            self.data_object_hash,
        )

    @cached_property
    def data_object_meta_file_path(self) -> Path:
        return self.data_object_dir / "meta"

    @cached_property
    def annotation_meta_dir_path(self) -> Path:
        return self.data_object_dir / "annotations"

    @cached_property
    def annotation_meta_file_path(self) -> Path:
        return self.annotation_meta_dir_path / self.annotation_hash

    @cached_property
    def annotation_dir(self) -> Path:
        return get_hash_path(
            self.ldb_dir / InstanceDir.ANNOTATIONS,
            self.annotation_hash,
        )

    @cached_property
    def data_object_hash(self) -> str:
        raise NotImplementedError

    @cached_property
    def annotation_hash(self) -> str:
        return hash_data(
            self.annotation_ldb_content_bytes + self.annotation_content_bytes,
        )

    @cached_property
    def data_object_meta(self) -> DataObjectMeta:
        raise NotImplementedError

    @cached_property
    def annotation_ldb_content(  # pylint: disable=no-self-use
        self,
    ) -> JSONObject:
        return {
            "user_version": None,
            "schema_version": None,
        }

    @cached_property
    def annotation_meta(self) -> AnnotationMeta:
        raise NotImplementedError

    @cached_property
    def annotation_content(self) -> JSONDecoded:
        raise NotImplementedError

    @cached_property
    def annotation_ldb_content_bytes(self) -> bytes:
        return json_dumps(self.annotation_ldb_content).encode()

    @cached_property
    def annotation_content_bytes(self) -> bytes:
        return json_dumps(self.annotation_content).encode()

    @cached_property
    def annotation_version(self) -> int:
        try:
            return len(list(self.annotation_meta_dir_path.iterdir())) + 1
        except FileNotFoundError:
            return 1

    @cached_property
    def has_annotation(self) -> bool:
        raise NotImplementedError

    def index_data(self) -> IndexedObjectResult:
        new_data_object = not self.data_object_dir.is_dir()
        self.enqueue_data(self.data_object_to_write())

        found_annotation = self.has_annotation
        new_annotation = False
        if found_annotation:
            new_annotation = not self.annotation_meta_file_path.is_file()
            self.enqueue_data(self.annotation_to_write())

        self.write_data()
        return IndexedObjectResult(
            found_data_object=True,
            found_annotation=found_annotation,
            new_data_object=new_data_object,
            new_annotation=new_annotation,
            data_object_hash=self.data_object_hash,
        )

    def data_object_to_write(self) -> List[DataToWrite]:
        return [
            (
                self.data_object_meta_file_path,
                json_dumps(self.data_object_meta).encode(),
                True,
            ),
        ]

    def annotation_to_write(self) -> List[DataToWrite]:
        annotation_meta_bytes = json_dumps(self.annotation_meta).encode()
        to_write = [
            (self.annotation_meta_file_path, annotation_meta_bytes, True),
        ]
        if not self.annotation_dir.is_dir():
            to_write.append(
                (
                    self.annotation_dir / "ldb",
                    self.annotation_ldb_content_bytes,
                    False,
                ),
            )
            to_write.append(
                (
                    self.annotation_dir / "user",
                    self.annotation_content_bytes,
                    False,
                ),
            )
        to_write.append(
            (
                self.data_object_dir / "current",
                self.annotation_hash.encode(),
                True,
            ),
        )
        return to_write

    def enqueue_data(self, data: Iterable[DataToWrite]) -> None:
        self._to_write.extend(data)

    def write_data(self) -> None:
        for file_path, data, overwrite_existing in self._to_write:
            write_data_file(file_path, data, overwrite_existing)


@dataclass
class AnnotationFileIndexingItem(IndexingItem):
    annotation_fsp: Optional[FileSystemPath]

    @cached_property
    def annotation_meta(self) -> AnnotationMeta:
        if self.annotation_fsp is None:
            raise IndexingException("Missing annotation_fsp")
        fs_info = self.annotation_fsp.fs.info(self.annotation_fsp.path)
        curr_mtime_float = fs_info.get("created")
        if curr_mtime_float is not None:
            curr_mtime: Optional[datetime] = timestamp_to_datetime(
                curr_mtime_float,
            )
        else:
            curr_mtime = None
        prev_annotation = (
            load_data_file(self.annotation_meta_file_path)
            if self.annotation_meta_file_path.exists()
            else {}
        )
        return construct_annotation_meta(
            prev_annotation,
            self.current_timestamp,
            self.annotation_version,
            curr_mtime,
        )

    @cached_property
    def annotation_content(self) -> JSONDecoded:
        if self.annotation_fsp is None:
            raise IndexingException("Missing annotation_fsp")
        return get_annotation_content(
            *self.annotation_fsp,
        )


@dataclass
class DataObjectFileIndexingItem(IndexingItem):
    data_object: FileSystemPath
    save_data_object_path_info: bool
    data_object_hash_cache: Mapping[AbstractFileSystem, Mapping[str, str]]

    @cached_property
    def data_object_hash(self) -> str:
        hash_str = None
        fs_hashes = self.data_object_hash_cache.get(self.data_object.fs)
        if fs_hashes is not None:
            hash_str = fs_hashes.get(self.data_object.path)
        if hash_str is None:
            hash_str = get_file_hash(
                self.data_object.fs,
                self.data_object.path,
            )
        return hash_str

    @cached_property
    def data_object_meta(self) -> DataObjectMeta:
        if not self.save_data_object_path_info:
            meta_contents: DataObjectMeta = load_data_file(
                self.data_object_meta_file_path,
            )
            meta_contents["last_indexed"] = self.current_timestamp
            meta_contents["tags"] = sorted(  # type: ignore[assignment]
                set(meta_contents["tags"])  # type: ignore[arg-type]
                | set(self.tags),
            )
        else:
            meta_contents = construct_data_object_meta(
                *self.data_object,
                load_data_file(self.data_object_meta_file_path)
                if self.data_object_meta_file_path.exists()
                else {},
                self.current_timestamp,
                tags=self.tags,
            )
        return meta_contents


@dataclass
class PairIndexingItem(
    AnnotationFileIndexingItem,
    DataObjectFileIndexingItem,
):
    @cached_property
    def data_object_dir(self) -> Path:
        return get_hash_path(
            self.ldb_dir / InstanceDir.DATA_OBJECT_INFO,
            self.data_object_hash,
        )

    @cached_property
    def data_object_meta_file_path(self) -> Path:
        return self.data_object_dir / "meta"

    @cached_property
    def has_annotation(self) -> bool:
        return self.annotation_fsp is not None



from dataclasses import dataclass
from pathlib import Path
from typing import Iterable, Optional, Tuple

from ldb.dataset import (
    get_collection,
    get_collection_dir_items,
    get_dataset,
    get_dataset_version_hash,
)
from ldb.path import InstanceDir, WorkspacePath
from ldb.utils import ROOT, parse_dataset_identifier
from ldb.workspace import load_workspace_dataset


@dataclass
class WorkspaceStatus:
    dataset_name: str
    dataset_version: int
    num_data_objects: int
    num_annotations: int


def status(ldb_dir: Path, dataset: str) -> WorkspaceStatus:
    item_gen: Iterable[Tuple[str, Optional[str]]]
    if not dataset:
        workspace_path = Path(".")
        workspace_ds = load_workspace_dataset(workspace_path)
        ds_name = workspace_ds.dataset_name
        ds_version = 0
        item_gen = get_collection_dir_items(
            workspace_path / WorkspacePath.COLLECTION,
            is_workspace=True,
        )
    else:
        ds_name, opt_ds_version = parse_dataset_identifier(dataset)
        ds_version = opt_ds_version or 0
        if ds_name == ROOT:
            item_gen = get_collection_dir_items(
                ldb_dir / InstanceDir.DATA_OBJECT_INFO,
                is_workspace=False,
            )
        else:
            dataset_obj = get_dataset(ldb_dir, ds_name)
            if not ds_version:
                ds_version = len(dataset_obj.versions)
            dataset_version_hash = get_dataset_version_hash(
                dataset_obj,
                ds_version,
            )
            item_gen = get_collection(ldb_dir, dataset_version_hash).items()
    num_data_objects = 0
    num_annotations = 0
    for _, annotation_hash in item_gen:
        num_data_objects += 1
        num_annotations += bool(annotation_hash)
    return WorkspaceStatus(
        dataset_name=ds_name,
        dataset_version=ds_version,
        num_data_objects=num_data_objects,
        num_annotations=num_annotations,
    )



import getpass
import os
from pathlib import Path

from ldb.dataset import CommitInfo, Dataset, DatasetVersion
from ldb.path import InstanceDir, WorkspacePath
from ldb.utils import (
    current_time,
    format_dataset_identifier,
    get_hash_path,
    hash_data,
    json_dumps,
    load_data_file,
    write_data_file,
)
from ldb.workspace import (
    collection_dir_to_object,
    load_workspace_dataset,
    workspace_dataset_is_clean,
)


def commit(
    ldb_dir: Path,
    workspace_path: Path,
    message: str,
) -> None:
    workspace_path = Path(os.path.normpath(workspace_path))
    workspace_ds = load_workspace_dataset(workspace_path)
    dataset_name = workspace_ds.dataset_name
    if workspace_dataset_is_clean(
        ldb_dir,
        workspace_ds,
        workspace_path,
    ):
        print("Nothing to commit.")
        return
    collection_obj = collection_dir_to_object(
        workspace_path / WorkspacePath.COLLECTION,
    )
    collection_obj_bytes = json_dumps(collection_obj).encode()
    collection_hash = hash_data(collection_obj_bytes)
    collection_path = get_hash_path(
        ldb_dir / InstanceDir.COLLECTIONS,
        collection_hash,
    )
    write_data_file(
        collection_path,
        collection_obj_bytes,
        overwrite_existing=False,
    )

    curr_time = current_time()
    username = getpass.getuser()

    dataset_file_path = ldb_dir / InstanceDir.DATASETS / dataset_name
    try:
        dataset = Dataset.parse(load_data_file(dataset_file_path))
    except FileNotFoundError:
        dataset = Dataset(
            name=dataset_name,
            created_by=username,
            created=curr_time,
            versions=[],
        )
    dataset_version = DatasetVersion(
        version=len(dataset.versions) + 1,
        parent=workspace_ds.parent,
        collection=collection_hash,
        tags=workspace_ds.tags.copy(),
        commit_info=CommitInfo(
            created_by=username,
            commit_time=curr_time,
            commit_message=message,
        ),
    )
    dataset_version_bytes = json_dumps(dataset_version.format()).encode()
    dataset_version_hash = hash_data(dataset_version_bytes)
    dataset_version_file_path = get_hash_path(
        ldb_dir / InstanceDir.DATASET_VERSIONS,
        dataset_version_hash,
    )
    write_data_file(dataset_version_file_path, dataset_version_bytes)
    dataset.versions.append(dataset_version_hash)
    write_data_file(
        dataset_file_path,
        json_dumps(dataset.format()).encode(),
        overwrite_existing=True,
    )
    workspace_ds.staged_time = curr_time
    workspace_ds.parent = dataset_version_hash
    write_data_file(
        workspace_path / WorkspacePath.DATASET,
        json_dumps(workspace_ds.format()).encode(),
        overwrite_existing=True,
    )
    dataset_identifier = format_dataset_identifier(
        dataset_name,
        dataset_version.version,
    )
    print(f"Committed {dataset_identifier}")



from dataclasses import dataclass
from pathlib import Path
from typing import Iterable, Iterator

from ldb.dataset import iter_datasets
from ldb.utils import format_dataset_identifier


@dataclass
class DSListing:
    name: str
    latest_version: int


def ds(
    ldb_dir: Path,
) -> Iterator[DSListing]:
    for dataset in iter_datasets(ldb_dir):
        yield DSListing(
            name=dataset.name,
            latest_version=len(dataset.versions),
        )


def print_ds_listings(ds_listings: Iterable[DSListing]) -> None:
    for item in ds_listings:
        print(format_dataset_identifier(item.name, item.latest_version))
